"""
This file contains utility functions used in repo scripts

For API calling, we support:
    - parallelized model calls on the same prompt 
    - batched model calls on different prompts

NOTE: any function beginning with 'a', e.g. 'afetch_together_api_response', is an asynchronous function
"""

import concurrent.futures
from typing import List, Mapping, Dict, Union, Optional, Literal, Any
import asyncio
import litellm
import pydantic
import pprint
import os
from dotenv import load_dotenv 

from judgeval.clients import async_together_client, together_client
from judgeval.constants import *
from judgeval.common.logger import debug, error

LITELLM_SUPPORTED_MODELS = set(litellm.model_list)

class CustomModelParameters(pydantic.BaseModel):
    model_name: str
    secret_key: str
    litellm_base_url: str
    
    @pydantic.field_validator('model_name')
    def validate_model_name(cls, v):
        if not v:
            raise ValueError("Model name cannot be empty")
        return v
    
    @pydantic.field_validator('secret_key')
    def validate_secret_key(cls, v):
        if not v:
            raise ValueError("Secret key cannot be empty")
        return v
    
    @pydantic.field_validator('litellm_base_url')
    def validate_litellm_base_url(cls, v):
        if not v:
            raise ValueError("Litellm base URL cannot be empty")
        return v

class ChatCompletionRequest(pydantic.BaseModel):
    model: str
    messages: List[Dict[str, str]]
    response_format: Optional[Union[pydantic.BaseModel, Dict[str, Any]]] = None
    
    @pydantic.field_validator('messages')
    def validate_messages(cls, messages):
        if not messages:
            raise ValueError("Messages cannot be empty")
            
        for msg in messages:
            if not isinstance(msg, dict):
                raise TypeError("Message must be a dictionary")
            if 'role' not in msg:
                raise ValueError("Message missing required 'role' field")
            if 'content' not in msg:
                raise ValueError("Message missing required 'content' field")
            if msg['role'] not in ['system', 'user', 'assistant']:
                raise ValueError(f"Invalid role '{msg['role']}'. Must be 'system', 'user', or 'assistant'")
        
        return messages

    @pydantic.field_validator('model')
    def validate_model(cls, model):
        if not model:
            raise ValueError("Model cannot be empty")
        if model not in TOGETHER_SUPPORTED_MODELS and model not in LITELLM_SUPPORTED_MODELS:
            raise ValueError(f"Model {model} is not in the list of supported models.")
        return model
    
    @pydantic.field_validator('response_format', mode='before')
    def validate_response_format(cls, response_format):
        if response_format is not None:
            if not isinstance(response_format, (dict, pydantic.BaseModel)):
                raise TypeError("Response format must be a dictionary or pydantic model")
            # Optional: Add additional validation for required fields if needed
            # For example, checking for 'type': 'json' in OpenAI's format
        return response_format

os.environ['LITELLM_LOG'] = 'DEBUG'

load_dotenv()

def read_file(file_path: str) -> str:
    with open(file_path, "r", encoding='utf-8') as file:
        return file.read()


def fetch_together_api_response(model: str, messages: List[Mapping], response_format: pydantic.BaseModel = None) -> str:
    """
    Fetches a single response from the Together API for a given model and messages.
    """
    # Validate request
    if messages is None or messages == []:
        raise ValueError("Messages cannot be empty")
    
    request = ChatCompletionRequest(
            model=model,
            messages=messages,
            response_format=response_format
    )
    
    debug(f"Calling Together API with model: {request.model}")
    debug(f"Messages: {request.messages}")
    
    if request.response_format is not None:
        debug(f"Using response format: {request.response_format}")
        response = together_client.chat.completions.create(
            model=TOGETHER_SUPPORTED_MODELS.get(request.model),
            messages=request.messages,
            response_format=request.response_format
        )
    else:
        response = together_client.chat.completions.create(
            model=TOGETHER_SUPPORTED_MODELS.get(request.model),
            messages=request.messages,
        )
    
    debug(f"Received response: {response.choices[0].message.content[:100]}...")
    return response.choices[0].message.content


async def afetch_together_api_response(model: str, messages: List[Mapping], response_format: pydantic.BaseModel = None) -> str:
    """
    ASYNCHRONOUSLY Fetches a single response from the Together API for a given model and messages.
    """
    request = ChatCompletionRequest(
        model=model,
        messages=messages,
        response_format=response_format
    )
    
    debug(f"Calling Together API with model: {request.model}")
    debug(f"Messages: {request.messages}")
    
    if request.response_format is not None:
        debug(f"Using response format: {request.response_format}")
        response = await async_together_client.chat.completions.create(
            model=TOGETHER_SUPPORTED_MODELS.get(request.model),
            messages=request.messages,
            response_format=request.response_format
        )
    else:
        response = await async_together_client.chat.completions.create(
            model=TOGETHER_SUPPORTED_MODELS.get(request.model),
            messages=request.messages,
        )
    return response.choices[0].message.content


def query_together_api_multiple_calls(models: List[str], messages: List[List[Mapping]], response_formats: List[pydantic.BaseModel] = None) -> List[str]:
    """
    Queries the Together API for multiple calls in parallel

    Args:
        models (List[str]): List of models to query
        messages (List[List[Mapping]]): List of messages to query. Each inner object corresponds to a single prompt.
        response_formats (List[pydantic.BaseModel], optional): A list of the format of the response if JSON forcing. Defaults to None.

    Returns:
        List[str]: TogetherAI responses for each model and message pair in order. Any exceptions in the thread call result in a None.
    """
    # Check for empty models list
    if not models:
        raise ValueError("Models list cannot be empty")

    # Validate all models are supported
    for model in models:
        if model not in TOGETHER_SUPPORTED_MODELS:
            raise ValueError(f"Model {model} is not in the list of supported TogetherAI models: {TOGETHER_SUPPORTED_MODELS}.")

    # Validate input lengths match
    if response_formats is None:
        response_formats = [None] * len(models)
    if not (len(models) == len(messages) == len(response_formats)):
        raise ValueError("Number of models, messages, and response formats must be the same")

    # Validate message format
    validate_batched_chat_messages(messages)

    num_workers = int(os.getenv('NUM_WORKER_THREADS', MAX_WORKER_THREADS))  
    # Initialize results to maintain ordered outputs
    out = [None] * len(messages)
    with concurrent.futures.ThreadPoolExecutor(max_workers=num_workers) as executor:
        # Submit all queries to together API with index, gets back the response content
        futures = {executor.submit(fetch_together_api_response, model, message, response_format): idx \
                   for idx, (model, message, response_format) in enumerate(zip(models, messages, response_formats))}
        
        # Collect results as they complete -- result is response content
        for future in concurrent.futures.as_completed(futures):
            idx = futures[future]
            try:
                out[idx] = future.result()
            except Exception as e:
                error(f"Error in parallel call {idx}: {str(e)}")
                out[idx] = None 
    return out


async def aquery_together_api_multiple_calls(models: List[str], messages: List[List[Mapping]], response_formats: List[pydantic.BaseModel] = None) -> List[str]:
    """
    Queries the Together API for multiple calls in parallel

    Args:
        models (List[str]): List of models to query
        messages (List[List[Mapping]]): List of messages to query. Each inner object corresponds to a single prompt.
        response_formats (List[pydantic.BaseModel], optional): A list of the format of the response if JSON forcing. Defaults to None.

    Returns:
        List[str]: TogetherAI responses for each model and message pair in order. Any exceptions in the thread call result in a None.
    """
    # Check for empty models list
    if not models:
        raise ValueError("Models list cannot be empty")

    # Validate all models are supported
    for model in models:
        if model not in TOGETHER_SUPPORTED_MODELS:
            raise ValueError(f"Model {model} is not in the list of supported TogetherAI models: {TOGETHER_SUPPORTED_MODELS}.")

    # Validate input lengths match
    if response_formats is None:
        response_formats = [None] * len(models)
    if not (len(models) == len(messages) == len(response_formats)):
        raise ValueError("Number of models, messages, and response formats must be the same")

    # Validate message format
    validate_batched_chat_messages(messages)

    debug(f"Starting parallel Together API calls for {len(messages)} messages")
    out = [None] * len(messages)
    
    async def fetch_and_store(idx, model, message, response_format):
        try:
            debug(f"Processing call {idx} with model {model}")
            out[idx] = await afetch_together_api_response(model, message, response_format)
        except Exception as e:
            error(f"Error in parallel call {idx}: {str(e)}")
            out[idx] = None

    tasks = [
        fetch_and_store(idx, model, message, response_format)
        for idx, (model, message, response_format) in enumerate(zip(models, messages, response_formats))
    ]
    
    await asyncio.gather(*tasks)
    debug(f"Completed {len(messages)} parallel calls")
    return out


def fetch_litellm_api_response(model: str, messages: List[Mapping], response_format: pydantic.BaseModel = None) -> str:
    """
    Fetches a single response from the Litellm API for a given model and messages.
    """
    request = ChatCompletionRequest(
        model=model,
        messages=messages,
        response_format=response_format
    )
    
    debug(f"Calling LiteLLM API with model: {request.model}")
    debug(f"Messages: {request.messages}")
    
    if request.response_format is not None:
        debug(f"Using response format: {request.response_format}")
        response = litellm.completion(
            model=request.model,
            messages=request.messages,
            response_format=request.response_format
        )
    else:
        response = litellm.completion(
            model=request.model,
            messages=request.messages,
        )
    return response.choices[0].message.content


def fetch_custom_litellm_api_response(custom_model_parameters: CustomModelParameters, messages: List[Mapping], response_format: pydantic.BaseModel = None) -> str:
    if messages is None or messages == []:
        raise ValueError("Messages cannot be empty")
    
    if custom_model_parameters is None:
        raise ValueError("Custom model parameters cannot be empty")
    
    if not isinstance(custom_model_parameters, CustomModelParameters):
        raise ValueError("Custom model parameters must be a CustomModelParameters object")
        
    if response_format is not None:
        response = litellm.completion(
            model=custom_model_parameters.model_name,
            messages=messages,
            api_key=custom_model_parameters.secret_key,
            base_url=custom_model_parameters.litellm_base_url,
            response_format=response_format
        )
    else:
        response = litellm.completion(
            model=custom_model_parameters.model_name,
            messages=messages,
            api_key=custom_model_parameters.secret_key,
            base_url=custom_model_parameters.litellm_base_url,
        )
    return response.choices[0].message.content


async def afetch_litellm_api_response(model: str, messages: List[Mapping], response_format: pydantic.BaseModel = None) -> str:
    """
    ASYNCHRONOUSLY Fetches a single response from the Litellm API for a given model and messages.
    """
    if messages is None or messages == []:
        raise ValueError("Messages cannot be empty")
    
    # Add validation
    validate_chat_messages(messages)
    
    if model not in LITELLM_SUPPORTED_MODELS:
        raise ValueError(f"Model {model} is not in the list of supported Litellm models: {LITELLM_SUPPORTED_MODELS}.")
    
    if response_format is not None:
        response = await litellm.acompletion(
            model=model,
            messages=messages,
            response_format=response_format
        )
    else:
        response = await litellm.acompletion(
            model=model,
            messages=messages,  
        )
    return response.choices[0].message.content


async def afetch_custom_litellm_api_response(custom_model_parameters: CustomModelParameters, messages: List[Mapping], response_format: pydantic.BaseModel = None) -> str:
    """
    ASYNCHRONOUSLY Fetches a single response from the Litellm API for a given model and messages.
    """
    if messages is None or messages == []:
        raise ValueError("Messages cannot be empty")
    
    if custom_model_parameters is None:
        raise ValueError("Custom model parameters cannot be empty")
    
    if not isinstance(custom_model_parameters, CustomModelParameters):
        raise ValueError("Custom model parameters must be a CustomModelParameters object")
        
    if response_format is not None:
        response = await litellm.acompletion(
            model=custom_model_parameters.model_name,
            messages=messages,
            api_key=custom_model_parameters.secret_key,
            base_url=custom_model_parameters.litellm_base_url,
            response_format=response_format
        )
    else:
        response = await litellm.acompletion(
            model=custom_model_parameters.model_name,
            messages=messages,
            api_key=custom_model_parameters.secret_key,
            base_url=custom_model_parameters.litellm_base_url,
        )
    return response.choices[0].message.content


def query_litellm_api_multiple_calls(models: List[str], messages: List[Mapping], response_formats: List[pydantic.BaseModel] = None) -> List[str]:
    """
    Queries the Litellm API for multiple calls in parallel

    Args:
        models (List[str]): List of models to query
        messages (List[Mapping]): List of messages to query
        response_formats (List[pydantic.BaseModel], optional): A list of the format of the response if JSON forcing. Defaults to None.

    Returns:
        List[str]: Litellm responses for each model and message pair in order. Any exceptions in the thread call result in a None.
    """
    num_workers = int(os.getenv('NUM_WORKER_THREADS', MAX_WORKER_THREADS))  
    # Initialize results to maintain ordered outputs
    out = [None] * len(messages)
    with concurrent.futures.ThreadPoolExecutor(max_workers=num_workers) as executor:
        # Submit all queries to Litellm API with index, gets back the response content
        futures = {executor.submit(fetch_litellm_api_response, model, message, response_format): idx \
                    for idx, (model, message, response_format) in enumerate(zip(models, messages, response_formats))}
        
        # Collect results as they complete -- result is response content
        for future in concurrent.futures.as_completed(futures):
            idx = futures[future]
            try:
                out[idx] = future.result()
            except Exception as e:
                error(f"Error in parallel call {idx}: {str(e)}")
                out[idx] = None 
    return out


async def aquery_litellm_api_multiple_calls(models: List[str], messages: List[Mapping], response_formats: List[pydantic.BaseModel] = None) -> List[str]:
    """
    Queries the Litellm API for multiple calls in parallel

    Args:
        models (List[str]): List of models to query
        messages (List[Mapping]): List of messages to query
        response_formats (List[pydantic.BaseModel], optional): A list of the format of the response if JSON forcing. Defaults to None.

    Returns:
        List[str]: Litellm responses for each model and message pair in order. Any exceptions in the thread call result in a None.
    """
    # Initialize results to maintain ordered outputs
    out = [None] * len(messages)
    
    async def fetch_and_store(idx, model, message, response_format):
        try:
            out[idx] = await afetch_litellm_api_response(model, message, response_format)
        except Exception as e:
            error(f"Error in parallel call {idx}: {str(e)}")
            out[idx] = None

    tasks = [
        fetch_and_store(idx, model, message, response_format)
        for idx, (model, message, response_format) in enumerate(zip(models, messages, response_formats))
    ]
    
    await asyncio.gather(*tasks)
    return out


def validate_chat_messages(messages, batched: bool = False):
    """Validate chat message format before API call"""
    if not isinstance(messages, list):
        raise TypeError("Messages must be a list")
    
    for msg in messages:
        if not isinstance(msg, dict):
            if batched and not isinstance(msg, list):
                raise TypeError("Each message must be a list")
            elif not batched:
                raise TypeError("Message must be a dictionary")
        if 'role' not in msg:
            raise ValueError("Message missing required 'role' field")
        if 'content' not in msg:
            raise ValueError("Message missing required 'content' field")
        if msg['role'] not in ['system', 'user', 'assistant']:
            raise ValueError(f"Invalid role '{msg['role']}'. Must be 'system', 'user', or 'assistant'")

def validate_batched_chat_messages(messages: List[List[Mapping]]):
    """
    Validate format of batched chat messages before API call
    
    Args:
        messages (List[List[Mapping]]): List of message lists, where each inner list contains
            message dictionaries with 'role' and 'content' fields
    
    Raises:
        TypeError: If messages format is invalid
        ValueError: If message content is invalid
    """
    if not isinstance(messages, list):
        raise TypeError("Batched messages must be a list")
    
    if not messages:
        raise ValueError("Batched messages cannot be empty")
        
    for message_list in messages:
        if not isinstance(message_list, list):
            raise TypeError("Each batch item must be a list of messages")
            
        # Validate individual messages using existing function
        validate_chat_messages(message_list)

def get_chat_completion(model_type: str, 
                        messages : Union[List[Mapping], List[List[Mapping]]], 
                        response_format: pydantic.BaseModel = None, 
                        batched: bool = False
                        ) -> Union[str, List[str]]:
    """
    Generates chat completions using a single model and potentially several messages. Supports closed-source and OSS models.

    Parameters:
        - model_type (str): The type of model to use for generating completions.
        - messages (Union[List[Mapping], List[List[Mapping]]]): The messages to be used for generating completions. 
            If batched is True, this should be a list of lists of mappings.
        - response_format (pydantic.BaseModel, optional): The format of the response. Defaults to None.
        - batched (bool, optional): Whether to process messages in batch mode. Defaults to False.
    Returns:
        - str: The generated chat completion(s). If batched is True, returns a list of strings.
    Raises:
        - ValueError: If requested model is not supported by Litellm or TogetherAI.
    """
    
    # Check for empty messages list
    if not messages or messages == []:
        raise ValueError("Messages cannot be empty")
    
    # Add validation
    if batched:
        validate_batched_chat_messages(messages)
    else:
        validate_chat_messages(messages)
    
    if batched and model_type in TOGETHER_SUPPORTED_MODELS:
        return query_together_api_multiple_calls(models=[model_type] * len(messages), 
                                                 messages=messages, 
                                                 response_formats=[response_format] * len(messages))
    elif batched and model_type in LITELLM_SUPPORTED_MODELS:
        return query_litellm_api_multiple_calls(models=[model_type] * len(messages), 
                                                messages=messages, 
                                                response_format=response_format)
    elif not batched and model_type in TOGETHER_SUPPORTED_MODELS:
        return fetch_together_api_response(model=model_type, 
                                           messages=messages, 
                                           response_format=response_format)
    elif not batched and model_type in LITELLM_SUPPORTED_MODELS:
        return fetch_litellm_api_response(model=model_type, 
                                          messages=messages, 
                                          response_format=response_format)
        
    
    
    raise ValueError(f"Model {model_type} is not supported by Litellm or TogetherAI for chat completions. Please check the model name and try again.")


async def aget_chat_completion(model_type: str, 
                               messages : Union[List[Mapping], List[List[Mapping]]], 
                               response_format: pydantic.BaseModel = None, 
                               batched: bool = False
                               ) -> Union[str, List[str]]:
    """
    ASYNCHRONOUSLY generates chat completions using a single model and potentially several messages. Supports closed-source and OSS models.

    Parameters:
        - model_type (str): The type of model to use for generating completions.
        - messages (Union[List[Mapping], List[List[Mapping]]]): The messages to be used for generating completions. 
            If batched is True, this should be a list of lists of mappings.
        - response_format (pydantic.BaseModel, optional): The format of the response. Defaults to None.
        - batched (bool, optional): Whether to process messages in batch mode. Defaults to False.
    Returns:
        - str: The generated chat completion(s). If batched is True, returns a list of strings.
    Raises:
        - ValueError: If requested model is not supported by Litellm or TogetherAI.
    """
    debug(f"Starting chat completion for model {model_type}, batched={batched}")
    
    if batched:
        validate_batched_chat_messages(messages)
    else:
        validate_chat_messages(messages)
    
    if batched and model_type in TOGETHER_SUPPORTED_MODELS:
        debug("Using batched Together API call")
        return await aquery_together_api_multiple_calls(models=[model_type] * len(messages), 
                                                        messages=messages, 
                                                        response_formats=[response_format] * len(messages))
    elif batched and model_type in LITELLM_SUPPORTED_MODELS:
        debug("Using batched LiteLLM API call")
        return await aquery_litellm_api_multiple_calls(models=[model_type] * len(messages), 
                                                       messages=messages, 
                                                       response_formats=[response_format] * len(messages))
    elif not batched and model_type in TOGETHER_SUPPORTED_MODELS:
        debug("Using single Together API call")
        return await afetch_together_api_response(model=model_type, 
                                                  messages=messages, 
                                                  response_format=response_format)
    elif not batched and model_type in LITELLM_SUPPORTED_MODELS:
        debug("Using single LiteLLM API call")
        return await afetch_litellm_api_response(model=model_type, 
                                                 messages=messages, 
                                                 response_format=response_format)
    
    error(f"Model {model_type} not supported by either API")
    raise ValueError(f"Model {model_type} is not supported by Litellm or TogetherAI for chat completions. Please check the model name and try again.")


def get_completion_multiple_models(models: List[str], messages: List[List[Mapping]], response_formats: List[pydantic.BaseModel] = None) -> List[str]:
    """
    Retrieves completions for a single prompt from multiple models in parallel. Supports closed-source and OSS models.

    Args:
        models (List[str]): List of models to query
        messages (List[List[Mapping]]): List of messages to query. Each inner object corresponds to a single prompt.
        response_formats (List[pydantic.BaseModel], optional): A list of the format of the response if JSON forcing. Defaults to None.
    
    Returns:
        List[str]: List of completions from the models in the order of the input models
    Raises:
        ValueError: If a model is not supported by Litellm or Together
    """
    debug(f"Starting multiple model completion for {len(models)} models")
    
    if models is None or models == []:
        raise ValueError("Models list cannot be empty")
    
    validate_batched_chat_messages(messages)
    
    if len(models) != len(messages):
        error(f"Model/message count mismatch: {len(models)} vs {len(messages)}")
        raise ValueError(f"Number of models and messages must be the same: {len(models)} != {len(messages)}")
    if response_formats is None:
        response_formats = [None] * len(models)
    # Partition the model requests into TogetherAI and Litellm models, but keep the ordering saved
    together_calls, litellm_calls = {}, {}  # index -> model, message, response_format
    together_responses, litellm_responses = [], []
    for idx, (model, message, r_format) in enumerate(zip(models, messages, response_formats)):
        if model in TOGETHER_SUPPORTED_MODELS:
            debug(f"Model {model} routed to Together API")
            together_calls[idx] = (model, message, r_format)
        elif model in LITELLM_SUPPORTED_MODELS:
            debug(f"Model {model} routed to LiteLLM API")
            litellm_calls[idx] = (model, message, r_format)
        else:
            error(f"Model {model} not supported by either API")
            raise ValueError(f"Model {model} is not supported by Litellm or TogetherAI for chat completions. Please check the model name and try again.")
    
    # Add validation before processing
    for msg_list in messages:
        validate_chat_messages(msg_list)
    
    # Get the responses from the TogetherAI models
    # List of responses from the TogetherAI models in order of the together_calls dict
    if together_calls:
        debug(f"Executing {len(together_calls)} Together API calls")
        together_responses = query_together_api_multiple_calls(models=[model for model, _, _ in together_calls.values()], 
                                                           messages=[message for _, message, _ in together_calls.values()], 
                                                           response_formats=[format for _, _, format in together_calls.values()])  
    
    # Get the responses from the Litellm models
    if litellm_calls:
        debug(f"Executing {len(litellm_calls)} LiteLLM API calls")
        litellm_responses = query_litellm_api_multiple_calls(models=[model for model, _, _ in litellm_calls.values()],
                                                        messages=[message for _, message, _ in litellm_calls.values()],
                                                        response_formats=[format for _, _, format in litellm_calls.values()])

    # Merge the responses in the order of the original models
    debug("Merging responses")
    out = [None] * len(models)
    for idx, (model, message, r_format) in together_calls.items():
        out[idx] = together_responses.pop(0)
    for idx, (model, message, r_format) in litellm_calls.items():
        out[idx] = litellm_responses.pop(0)
    debug("Multiple model completion finished")
    return out 


async def aget_completion_multiple_models(models: List[str], messages: List[List[Mapping]], response_formats: List[pydantic.BaseModel] = None) -> List[str]:
    """
    ASYNCHRONOUSLY retrieves completions for a single prompt from multiple models in parallel. Supports closed-source and OSS models.

    Args:
        models (List[str]): List of models to query
        messages (List[List[Mapping]]): List of messages to query. Each inner object corresponds to a single prompt.
        response_formats (List[pydantic.BaseModel], optional): A list of the format of the response if JSON forcing. Defaults to None.
    
    Returns:
        List[str]: List of completions from the models in the order of the input models
    Raises:
        ValueError: If a model is not supported by Litellm or Together
    """
    if models is None or models == []:
        raise ValueError("Models list cannot be empty")
    
    if len(models) != len(messages):
        raise ValueError(f"Number of models and messages must be the same: {len(models)} != {len(messages)}")
    if response_formats is None:
        response_formats = [None] * len(models)

    validate_batched_chat_messages(messages)
    
    # Partition the model requests into TogetherAI and Litellm models, but keep the ordering saved
    together_calls, litellm_calls = {}, {}  # index -> model, message, response_format
    together_responses, litellm_responses = [], []
    for idx, (model, message, r_format) in enumerate(zip(models, messages, response_formats)):
        if model in TOGETHER_SUPPORTED_MODELS:
            together_calls[idx] = (model, message, r_format)
        elif model in LITELLM_SUPPORTED_MODELS:
            litellm_calls[idx] = (model, message, r_format)
        else:
            raise ValueError(f"Model {model} is not supported by Litellm or TogetherAI for chat completions. Please check the model name and try again.")
    
    # Add validation before processing
    for msg_list in messages:
        validate_chat_messages(msg_list)
    
    # Get the responses from the TogetherAI models
    # List of responses from the TogetherAI models in order of the together_calls dict
    if together_calls:
        together_responses = await aquery_together_api_multiple_calls(
            models=[model for model, _, _ in together_calls.values()], 
            messages=[message for _, message, _ in together_calls.values()], 
            response_formats=[format for _, _, format in together_calls.values()]
        )
    
    # Get the responses from the Litellm models
    if litellm_calls:
        litellm_responses = await aquery_litellm_api_multiple_calls(
            models=[model for model, _, _ in litellm_calls.values()],
            messages=[message for _, message, _ in litellm_calls.values()],
            response_formats=[format for _, _, format in litellm_calls.values()]
        )

    # Merge the responses in the order of the original models
    out = [None] * len(models)
    for idx, (model, message, r_format) in together_calls.items():
        out[idx] = together_responses.pop(0)
    for idx, (model, message, r_format) in litellm_calls.items():
        out[idx] = litellm_responses.pop(0)
    return out


if __name__ == "__main__":
    
    # Batched
    pprint.pprint(get_chat_completion(
        model_type="LLAMA3_405B_INSTRUCT_TURBO",
        messages=[
            [
                {"role": "system", "content": "You are a helpful assistant."},
                {"role": "user", "content": "What is the capital of France?"},
            ],
            [
                {"role": "system", "content": "You are a helpful assistant."},
                {"role": "user", "content": "What is the capital of Japan?"},
            ]
        ],
        batched=True
    ))

    # Non batched
    pprint.pprint(get_chat_completion(
        model_type="LLAMA3_8B_INSTRUCT_TURBO",
        messages=[
            {"role": "system", "content": "You are a helpful assistant."},
            {"role": "user", "content": "What is the capital of France?"},
        ],
        batched=False
    ))

    # Batched single completion to multiple models
    pprint.pprint(get_completion_multiple_models(
        models=[
            "LLAMA3_70B_INSTRUCT_TURBO", "LLAMA3_405B_INSTRUCT_TURBO", "gpt-4o-mini"
        ],
        messages=[
            [
                {"role": "system", "content": "You are a helpful assistant."},
                {"role": "user", "content": "What is the capital of China?"},
            ],
            [
                {"role": "system", "content": "You are a helpful assistant."},
                {"role": "user", "content": "What is the capital of France?"},
            ],
            [
                {"role": "system", "content": "You are a helpful assistant."},
                {"role": "user", "content": "What is the capital of Japan?"},
            ]
        ]
    ))



from typing import List, Optional
from pydantic import BaseModel, Field


class HallucinationVerdict(BaseModel):
    verdict: str
    reason: str


class Verdicts(BaseModel):
    verdicts: List[HallucinationVerdict]


class Reason(BaseModel):
    reason: str


class HallucinationTemplate:
    @staticmethod
    def generate_verdicts(actual_output, contexts):
        return f"""==== TASK INSTRUCTIONS ====
You will be provided with an `actual output` (the response of an LLM to a particular query) and `contexts` (ground truth contextual information from a knowledge base).
Your task is to take each context in contexts and determine whether the `actual output` factually agrees with the context.

Additional notes:
You should NOT use any prior knowledge you have in your decision making process; take each context at face value. 
Since you will determine a verdict for EACH context, the number of 'verdicts' is EXACTLY EQUAL TO the number of contexts. 
You should be lenient in your judgment when the actual output lacks detail with respect to the context segment; you should ONLY provide a 'no' answer if the context contradicts the actual output.

==== FORMATTING INSTRUCTIONS ====
You should return a JSON object with a key 'verdicts', which is a list of JSON objects. Each JSON object corresponds to a context in `contexts`, and should have 2 fields: 'verdict' and 'reason'. 
The 'verdict' key should be EXACTLY one of 'yes' or 'no', representing whether the `actual output` factually agrees with the context segment. 
The 'reason' is the justification for the verdict. If your verdict is 'no', try to provide a correction in the reason. 

==== EXAMPLE ====
Example contexts: ["Einstein won the Nobel Prize for his discovery of the photoelectric effect.", "Einstein won the Nobel Prize in 1968."]
Example actual output: "Einstein won the Nobel Prize in 1969 for his discovery of the photoelectric effect."

Example:
{{
    "verdicts": [
        {{
            "verdict": "yes",
            "reason": "The actual output agrees with the provided context which states that Einstein won the Nobel Prize for his discovery of the photoelectric effect."
        }},
        {{
            "verdict": "no",
            "reason": "The actual output contradicts the provided context which states that Einstein won the Nobel Prize in 1968, not 1969."
        }}
    ]  
}}

==== YOUR TURN ====
Contexts:
{contexts}

Actual Output:
{actual_output}

JSON:
"""

    @staticmethod
    def generate_reason(contradictions, score):
        return f"""==== TASK INSTRUCTIONS ====
An LLM has been provided with a list of `contexts` (ground truth contextual information from a knowledge base) and `actual output` (the response of an LLM to a particular query). 
You will be provided with a list of `contradictions`, which are factual discrepancies between the context segments and the actual output. 
Additionally, you will be provided with a hallucination score, which is a float (0 - 1, where 0 is the best score) indicating the fraction of context segments that contradict the actual output.

Your task is to provide a CLEAR and CONCISE reason for the hallucination score. 
If the hallucination score is 0 (no contradictions), you should instead respond with a positive remark with an upbeat encouraging tone (but don't overblow the kind attitude).
        
==== FORMATTING INSTRUCTIONS ====
Please make sure to only return in JSON format, with the 'reason' key providing the reason.
Example JSON:
{{
    "reason": "The score is <hallucination_score> because <your_reason>."
}}

==== EXAMPLE ====
Example Contradictions:
[
    "The actual output claims Einstein won the Nobel Prize in 1969, which contradicts the context stating he won it in 1968.",
    "The actual output states Einstein was a chemist, but the context indicates he was a physicist.",
    "The actual output claims Einstein was born in Switzerland, while the context states he was born in Germany."
]

Example Hallucination Score: 
0.75

Example Response:
{{
    "reason": "The score is 0.75 because the actual output made multiple factual errors: incorrectly stating Einstein's Nobel Prize year (1969 vs 1968), his profession (chemist vs physicist), and birthplace (Switzerland vs Germany)."
}}

==== YOUR TURN ====
Contradictions:
{contradictions}

Hallucination Score:
{score}

JSON:
"""



"""
ClassifierScorer implementation for basic Text-to-SQL evaluation.

Takes a natural language query, a corresponding LLM-generated SQL query, and a table schema + (optional) metadata.
Determines if the LLM-generated SQL query is valid and works for the natural language query.
"""
from judgeval.scorers import ClassifierScorer

Text2SQLScorer = ClassifierScorer(
    "Text to SQL",
    slug="text2sql-1010101010",
    threshold=1.0,
    conversation=[{
        "role": "system",
        "content": """You will be given a natural language query, a corresponding LLM-generated SQL query, and a table schema + (optional) metadata.

** TASK INSTRUCTIONS **
Your task is to decide whether the LLM generated SQL query properly filters for what the natural language query is asking, based on the table schema + (optional) metadata. 
Additionally, you should check if the SQL query is valid based on the table schema (checking for syntax errors, false column names, etc.)

** TIPS **
- Look for correct references to the table schema for column names, table names, etc.
- Check that the SQL query can be executed; make sure JOINs, GROUP BYs, ORDER BYs, etc. are valid with respect to the table schema.
- Check that aggregation functions (COUNT, SUM, AVG, etc.) are used appropriately with GROUP BY clauses
- Verify that WHERE conditions use the correct operators and data types for comparisons
- Ensure LIMIT and OFFSET clauses make sense for the query's purpose
- Check that JOINs use the correct keys and maintain referential integrity
- Verify that ORDER BY clauses use valid column names and sort directions
- Check for proper handling of NULL values where relevant
- Ensure subqueries are properly constructed and correlated when needed
- EVEN IF THE QUERY IS VALID, IF IT DOESN'T WORK FOR THE NATURAL LANGUAGE QUERY, YOU SHOULD CHOOSE "N" AS THE ANSWER.

** FORMATTING YOUR ANSWER **
If the SQL query is valid and works for the natural language query, choose option "Y" and otherwise "N". Provide a justification for your decision; if you choose "N", explain what about the LLM-generated SQL query is incorrect, or explain why it doesn't address the natural language query. 
IF YOUR JUSTIFICATION SHOWS THAT THE SQL QUERY IS VALID AND WORKS FOR THE NATURAL LANGUAGE QUERY, YOU SHOULD CHOOSE "Y" AS THE ANSWER. 
IF THE SQL QUERY IS INVALID, YOU SHOULD CHOOSE "N" AS THE ANSWER. 

** YOUR TURN **
Natural language query:
{{input}}

LLM generated SQL query:
{{actual_output}}

Table schema:
{{context}}
        """
    }],
    options={
        "Y": 1.0, 
        "N": 0.0
    }
)




import openai
import os
import asyncio
from tavily import TavilyClient
from dotenv import load_dotenv
import chromadb
from chromadb.utils import embedding_functions

from judgeval.common.tracer import Tracer, wrap
from judgeval.scorers import FaithfulnessScorer, AnswerRelevancyScorer
from demo.cookbooks.openai_travel_agent.populate_db import destinations_data

judgment = Tracer(api_key=os.getenv("JUDGMENT_API_KEY"))


def populate_vector_db(collection, destinations_data):
    """
    Populate the vector DB with travel information.
    destinations_data should be a list of dictionaries with 'destination' and 'information' keys
    """
    for data in destinations_data:
        collection.add(
            documents=[data['information']],
            metadatas=[{"destination": data['destination']}],
            ids=[f"destination_{data['destination'].lower().replace(' ', '_')}"]
        )

@judgment.observe(span_type="search tool")
def search_tavily(query):
    """Fetch travel data using Tavily API."""
    API_KEY = os.getenv("TAVILY_API_KEY")
    client = TavilyClient(api_key=API_KEY)
    results = client.search(query, num_results=3)
    return results

@judgment.observe(span_type="tool")
async def get_attractions(destination):
    """Search for top attractions in the destination."""
    prompt = f"Best tourist attractions in {destination}"
    attractions_search = search_tavily(prompt)
    await judgment.get_current_trace().async_evaluate(
        scorers=[AnswerRelevancyScorer(threshold=0.5)],
        input=prompt,
        actual_output=str(attractions_search),
        model="gpt-4o-mini",
        log_results=True
    )
    return attractions_search

@judgment.observe(span_type="tool")
async def get_hotels(destination):
    """Search for hotels in the destination."""
    prompt = f"Best hotels in {destination}"
    hotels_search = search_tavily(prompt)
    await judgment.get_current_trace().async_evaluate(
        scorers=[AnswerRelevancyScorer(threshold=0.5)],
        input=prompt,
        actual_output=str(hotels_search),
        model="gpt-4o-mini",
        log_results=True
    )
    return hotels_search

@judgment.observe(span_type="tool")
async def get_flights(destination):
    """Search for flights to the destination."""
    prompt = f"Flights to {destination} from major cities"
    flights_search = search_tavily(prompt)
    await judgment.get_current_trace().async_evaluate(
        scorers=[AnswerRelevancyScorer(threshold=0.5)],
        input=prompt,
        actual_output=str(flights_search),
        model="gpt-4o-mini",
        log_results=True
    )
    return flights_search

@judgment.observe(span_type="tool")
async def get_weather(destination, start_date, end_date):
    """Search for weather information."""
    prompt = f"Weather forecast for {destination} from {start_date} to {end_date}"
    weather_search = search_tavily(prompt)
    # await judgment.get_current_trace().async_evaluate(
    #     scorers=[AnswerRelevancyScorer(threshold=0.5)],
    #     input=prompt,
    #     actual_output=str(weather_search),
    #     model="gpt-4o-mini",
    #     log_results=True
    # )
    return weather_search

@judgment.observe(span_type="Retriever")
def initialize_vector_db():
    """Initialize ChromaDB with OpenAI embeddings."""
    client = chromadb.Client()
    embedding_fn = embedding_functions.OpenAIEmbeddingFunction(
        api_key=os.getenv("OPENAI_API_KEY"),
        model_name="text-embedding-3-small"
    )
    res = client.get_or_create_collection(
        "travel_information",
        embedding_function=embedding_fn
    )
    populate_vector_db(res, destinations_data)
    return res

@judgment.observe(span_type="Retriever")
def query_vector_db(collection, destination, k=3):
    """Query the vector database for existing travel information."""
    try:
        results = collection.query(
            query_texts=[destination],
            n_results=k
        )
        return results['documents'][0] if results['documents'] else []
    except Exception:
        return []

@judgment.observe(span_type="function")
async def research_destination(destination, start_date, end_date):
    """Gather all necessary travel information for a destination."""
    # First, check the vector database
    collection = initialize_vector_db()
    existing_info = query_vector_db(collection, destination)
    
    # Get real-time information from Tavily
    tavily_data = {
        "attractions": await get_attractions(destination),
        "hotels": await get_hotels(destination),
        "flights": await get_flights(destination),
        "weather": await get_weather(destination, start_date, end_date)
    }
    
    return {
        "vector_db_results": existing_info,
        **tavily_data
    }

@judgment.observe(span_type="function")
async def create_travel_plan(destination, start_date, end_date, research_data):
    """Generate a travel itinerary using the researched data."""
    vector_db_context = "\n".join(research_data['vector_db_results']) if research_data['vector_db_results'] else "No pre-stored information available."
    
    prompt = f"""
    Create a structured travel itinerary for a trip to {destination} from {start_date} to {end_date}.
    
    Pre-stored destination information:
    {vector_db_context}
    
    Current travel data:
    - Attractions: {research_data['attractions']}
    - Hotels: {research_data['hotels']}
    - Flights: {research_data['flights']}
    - Weather: {research_data['weather']}
    """
    
    client = wrap(openai.Client(api_key=os.getenv("OPENAI_API_KEY")))
    response = client.chat.completions.create(
        model="gpt-4",
        messages=[
            {"role": "system", "content": "You are an expert travel planner. Combine both historical and current information to create the best possible itinerary."},
            {"role": "user", "content": prompt}
        ]
    ).choices[0].message.content
    

    await judgment.get_current_trace().async_evaluate(
        scorers=[
            FaithfulnessScorer(threshold=0.5)
        ],
        input="",
        actual_output=response,
        retrieval_context=[
            vector_db_context, 
            str(research_data['attractions']), 
            str(research_data['hotels']), 
            str(research_data['flights']), 
            str(research_data['weather'])
        ],
        model="gpt-4o-mini",
        log_results=True
    )
    return response


async def generate_itinerary(destination, start_date, end_date):
    """Main function to generate a travel itinerary."""
    with judgment.trace(
        "generate_itinerary_demo",
        project_name="travel_agent_demo",
        overwrite=True
    ) as trace:    
        research_data = await research_destination(destination, start_date, end_date)
        res = await create_travel_plan(destination, start_date, end_date, research_data)
        trace.save()
        trace.print()
        return res


if __name__ == "__main__":
    load_dotenv()
    destination = input("Enter your travel destination: ")
    start_date = input("Enter start date (YYYY-MM-DD): ")
    end_date = input("Enter end date (YYYY-MM-DD): ")
    itinerary = asyncio.run(generate_itinerary(destination, start_date, end_date))
    print("\nGenerated Itinerary:\n", itinerary)


import pydantic
from typing import List, Union, Mapping

from judgeval import *
from judgeval.judges import JudgevalJudge
from judgeval.common.utils import afetch_litellm_api_response, fetch_litellm_api_response
from judgeval.common.logger import debug, error

BASE_CONVERSATION = [
    {"role": "system", "content": "You are a helpful assistant."},
]  # for string inputs, we need to add the user query to a base conversation, since LiteLLM only accepts a list of dictionaries as a chat history


class LiteLLMJudge(JudgevalJudge):
    def __init__(self, model: str = "gpt-4o-mini", **kwargs):
        debug(f"Initializing LiteLLMJudge with model={model}")
        self.model = model
        self.kwargs = kwargs
        super().__init__(model_name=model)

    def generate(self, input: Union[str, List[Mapping[str, str]]], schema: pydantic.BaseModel = None) -> str:
        debug(f"Generating response for input type: {type(input)}")
        if isinstance(input, str):
            convo = BASE_CONVERSATION + [{"role": "user", "content": input}]
            return fetch_litellm_api_response(model=self.model, messages=convo, response_format=schema)
        elif isinstance(input, list):
            return fetch_litellm_api_response(model=self.model, messages=input, response_format=schema)
        else:
            error(f"Invalid input type received: {type(input)}")
            raise TypeError(f"Input must be a string or a list of dictionaries. Input type of: {type(input)}")

    async def a_generate(self, input: Union[str, List[Mapping[str, str]]], schema: pydantic.BaseModel = None) -> str:
        debug(f"Async generating response for input type: {type(input)}")
        if isinstance(input, str):
            convo = BASE_CONVERSATION + [{"role": "user", "content": input}]
            response = await afetch_litellm_api_response(model=self.model, messages=convo, response_format=schema)
            return response
        elif isinstance(input, list):
            response = await afetch_litellm_api_response(model=self.model, messages=input, response_format=schema)
            return response
        else:
            error(f"Invalid input type received: {type(input)}")
            raise TypeError(f"Input must be a string or a list of dictionaries. Input type of: {type(input)}")
    
    def load_model(self):
        return self.model

    def get_model_name(self) -> str:
        return self.model



from typing import Optional, List, Union
import asyncio

from judgeval.constants import APIScorer
from judgeval.scorers.utils import (get_or_create_event_loop,
                                    scorer_progress_meter,
                                    create_verbose_logs,
                                    parse_response_json,
                                    check_example_params
                                    )
from judgeval.scorers import JudgevalScorer
from judgeval.judges import JudgevalJudge
from judgeval.judges.utils import create_judge
from judgeval.data import Example, ExampleParams
from judgeval.scorers.judgeval_scorers.local_implementations.contextual_relevancy.prompts import *


required_params = [
    ExampleParams.INPUT,
    ExampleParams.ACTUAL_OUTPUT,
    ExampleParams.RETRIEVAL_CONTEXT,
]


class ContextualRelevancyScorer(JudgevalScorer):
    def __init__(
        self,
        threshold: float = 0.5,
        model: Optional[Union[str, JudgevalJudge]] = None,
        include_reason: bool = True,
        async_mode: bool = True,
        strict_mode: bool = False,
        verbose_mode: bool = False,
        user: Optional[str] = None
    ):
        super().__init__(
            score_type=APIScorer.CONTEXTUAL_RELEVANCY,
            threshold=1 if strict_mode else threshold,
            evaluation_model=None,
            include_reason=include_reason,
            async_mode=async_mode,
            strict_mode=strict_mode,
            verbose_mode=verbose_mode
        )
        self.user = user
        self.model, self.using_native_model = create_judge(model)
        self.evaluation_model = self.model.get_model_name()

    def score_example(
        self,
        example: Example,
        _show_indicator: bool = True,
    ) -> float:
        check_example_params(example, required_params, self)
        
        with scorer_progress_meter(self, display_meter=_show_indicator):
            if self.async_mode:
                loop = get_or_create_event_loop()
                loop.run_until_complete(
                    self.a_score_example(example, _show_indicator=False)
                )
            else:
                self.verdicts_list: List[ContextualRelevancyVerdicts] = [
                    (self._generate_verdicts(example.input, context))
                    for context in example.retrieval_context
                ]
                self.score = self._calculate_score()
                self.reason = self._generate_reason(example.input)
                self.success = self.score >= self.threshold
                self.verbose_logs = create_verbose_logs(
                    self,
                    steps=[
                        f"Verdicts:\n{[v.model_dump() for v in self.verdicts_list]}",
                        f"Score: {self.score}\nReason: {self.reason}",
                    ],
                )

                return self.score

    async def a_score_example(
        self,
        example: Example,
        _show_indicator: bool = True,
    ) -> float:
        check_example_params(example, required_params, self)
        
        with scorer_progress_meter(
            self,
            async_mode=True,
            display_meter=_show_indicator,
        ):
            self.verdicts_list: List[ContextualRelevancyVerdicts] = (
                await asyncio.gather(
                    *[
                        self._a_generate_verdicts(example.input, context)
                        for context in example.retrieval_context
                    ]
                )
            )
            self.score = self._calculate_score()
            self.reason = await self._a_generate_reason(example.input)
            self.success = self.score >= self.threshold
            self.verbose_logs = create_verbose_logs(
                self,
                steps=[
                    f"Verdicts:\n{[v.model_dump() for v in self.verdicts_list]}",
                    f"Score: {self.score}\nReason: {self.reason}",
                ],
            )
            return self.score

    async def _a_generate_reason(self, input: str):
        if self.include_reason is False:
            return None

        irrelevancies = []
        relevant_statements = []
        for verdicts in self.verdicts_list:
            for verdict in verdicts.verdicts:
                if verdict.verdict.lower() == "no":
                    irrelevancies.append(verdict.model_dump())
                else:
                    relevant_statements.append(verdict.model_dump())
        prompt: dict = ContextualRelevancyTemplate.generate_reason(
            input=input,
            irrelevancies=irrelevancies,
            relevant_statements=relevant_statements,
            score=format(self.score, ".2f"),
        )
        if self.using_native_model:
            res = await self.model.a_generate(prompt)
            data = parse_response_json(res, self)
            return data["reason"]
        else:
            try:
                res: Reason = await self.model.a_generate(prompt, schema=Reason)
                return res.reason
            except TypeError:
                res = await self.model.a_generate(prompt)
                data = parse_response_json(res, self)
                return data["reason"]

    def _generate_reason(self, input: str):
        if self.include_reason is False:
            return None

        irrelevancies = []
        relevant_statements = []
        for verdicts in self.verdicts_list:
            for verdict in verdicts.verdicts:
                if verdict.verdict.lower() == "no":
                    irrelevancies.append(verdict.reason)
                else:
                    relevant_statements.append(verdict.statement)

        prompt: dict = ContextualRelevancyTemplate.generate_reason(
            input=input,
            irrelevancies=irrelevancies,
            relevant_statements=relevant_statements,
            score=format(self.score, ".2f"),
        )
        if self.using_native_model:
            res = self.model.generate(prompt)
            data = parse_response_json(res, self)
            return data["reason"]
        else:
            try:
                res: Reason = self.model.generate(prompt, schema=Reason)
                return res.reason
            except TypeError:
                res = self.model.generate(prompt)
                data = parse_response_json(res, self)
                return data["reason"]

    def _calculate_score(self):
        total_verdicts = 0
        relevant_statements = 0
        for verdicts in self.verdicts_list:
            for verdict in verdicts.verdicts:
                total_verdicts += 1
                if verdict.verdict.lower() == "yes":
                    relevant_statements += 1

        if total_verdicts == 0:
            return 0

        score = relevant_statements / total_verdicts
        return 0 if self.strict_mode and score < self.threshold else score

    async def _a_generate_verdicts(
        self, input: str, context: List[str]
    ) -> ContextualRelevancyVerdicts:
        prompt = ContextualRelevancyTemplate.generate_verdicts(
            input=input, context=context
        )
        if self.using_native_model:
            res = await self.model.a_generate(prompt)
            data = parse_response_json(res, self)
            return ContextualRelevancyVerdicts(**data)
        else:
            try:
                res = await self.model.a_generate(
                    prompt, schema=ContextualRelevancyVerdicts
                )
                return res
            except TypeError:
                res = await self.model.a_generate(prompt)
                data = parse_response_json(res, self)
                return ContextualRelevancyVerdicts(**data)

    def _generate_verdicts(
        self, input: str, context: str
    ) -> ContextualRelevancyVerdicts:
        prompt = ContextualRelevancyTemplate.generate_verdicts(
            input=input, context=context
        )
        if self.using_native_model:
            res = self.model.generate(prompt)
            data = parse_response_json(res, self)
            return ContextualRelevancyVerdicts(**data)
        else:
            try:
                res = self.model.generate(
                    prompt, schema=ContextualRelevancyVerdicts
                )
                return res
            except TypeError:
                res = self.model.generate(prompt)
                data = parse_response_json(res, self)
                return ContextualRelevancyVerdicts(**data)

    def _success_check(self) -> bool:
        if self.error is not None:
            self.success = False
        else:
            try:
                self.success = self.score >= self.threshold
            except:
                self.success = False
        return self.success

    @property
    def __name__(self):
        return "Contextual Relevancy"
    


"""
`judgeval` tool correctness scorer

TODO add link to docs page for this scorer

"""

# Internal imports
from judgeval.scorers.api_scorer import APIJudgmentScorer
from judgeval.constants import APIScorer


class ToolCorrectnessScorer(APIJudgmentScorer):
    def __init__(self, threshold: float):
        super().__init__(threshold=threshold, score_type=APIScorer.TOOL_CORRECTNESS)

    @property
    def __name__(self):
        return "Tool Correctness"



import asyncio
import requests
from typing import List, Dict
from datetime import datetime
from rich import print as rprint

from judgeval.data import (
    Example, 
    ScorerData, 
    ScoringResult
)
from judgeval.scorers import (
    JudgevalScorer, 
    APIJudgmentScorer,
    ClassifierScorer
)
from judgeval.scorers.score import a_execute_scoring

from judgeval.constants import (
    ROOT_API,
    JUDGMENT_EVAL_API_URL,
    JUDGMENT_EVAL_LOG_API_URL,
)
from judgeval.common.exceptions import JudgmentAPIError
from judgeval.evaluation_run import EvaluationRun
from judgeval.common.logger import (
    enable_logging, 
    debug, 
    info, 
    error, 
    example_logging_context
)


def execute_api_eval(evaluation_run: EvaluationRun) -> List[Dict]:
    """
    Executes an evaluation of a list of `Example`s using one or more `JudgmentScorer`s via the Judgment API.

    Args:
        evaluation_run (EvaluationRun): The evaluation run object containing the examples, scorers, and metadata

    Returns:
        List[Dict]: The results of the evaluation. Each result is a dictionary containing the fields of a `ScoringResult`
                    object. 
    """
    
    try:
        # submit API request to execute evals
        payload = evaluation_run.model_dump(warnings=False)
        response = requests.post(JUDGMENT_EVAL_API_URL, json=payload)
        response_data = response.json()
    except Exception as e:
        error(f"Error: {e}")
        details = response.json().get("detail", "No details provided")
        raise JudgmentAPIError("An error occurred while executing the Judgment API request: " + details)
    # Check if the response status code is not 2XX
    # Add check for the duplicate eval run name
    if not response.ok:
        error_message = response_data.get('detail', 'An unknown error occurred.')
        error(f"Error: {error_message=}")
        raise JudgmentAPIError(error_message)
    return response_data


def merge_results(api_results: List[ScoringResult], local_results: List[ScoringResult]) -> List[ScoringResult]:
    """
    When executing scorers that come from both the Judgment API and local scorers, we're left with
    results for each type of scorer. This function merges the results from the API and local evaluations,
    grouped by example. In particular, we merge the `scorers_data` field of each `ScoringResult` object.

    Args:
        api_results (List[ScoringResult]): The `ScoringResult`s from the API evaluation
        local_results (List[ScoringResult]): The `ScoringResult`s from the local evaluation

    Returns:
        List[ScoringResult]: The merged `ScoringResult`s (updated `scorers_data` field)
    """
    # No merge required
    if not local_results and api_results:
        return api_results
    if not api_results and local_results:
        return local_results

    if len(api_results) != len(local_results):
        # Results should be of same length because each ScoringResult is a 1-1 mapping to an Example
        raise ValueError(f"The number of API and local results do not match: {len(api_results)} vs {len(local_results)}")
    
    # Each ScoringResult in api and local have all the same fields besides `scorers_data`
    for api_result, local_result in zip(api_results, local_results):
        if api_result.input != local_result.input:
            raise ValueError("The API and local results are not aligned.")
        if api_result.actual_output != local_result.actual_output:
            raise ValueError("The API and local results are not aligned.")
        if api_result.expected_output != local_result.expected_output:
            raise ValueError("The API and local results are not aligned.")
        if api_result.context != local_result.context:
            raise ValueError("The API and local results are not aligned.")
        if api_result.retrieval_context != local_result.retrieval_context:
            raise ValueError("The API and local results are not aligned.")
        
        # Merge ScorerData from the API and local scorers together
        api_scorer_data = api_result.scorers_data
        local_scorer_data = local_result.scorers_data
        if api_scorer_data is None and local_scorer_data is not None:
            api_result.scorers_data = local_scorer_data

        if api_scorer_data is not None and local_scorer_data is not None:
            api_result.scorers_data = api_scorer_data + local_scorer_data
    
    return api_results


def check_missing_scorer_data(results: List[ScoringResult]) -> List[ScoringResult]:
    """
    Checks if any `ScoringResult` objects are missing `scorers_data`.

    If any are missing, logs an error and returns the results.
    """
    for i, result in enumerate(results):
        if not result.scorers_data:
            error(
                f"Scorer data is missing for example {i}. "
                "This is usually caused when the example does not contain "
                "the fields required by the scorer. "
                "Check that your example contains the fields required by the scorers. "
                "TODO add docs link here for reference."
            )
    return results


def check_eval_run_name_exists(eval_name: str, project_name: str, judgment_api_key: str) -> None:
    """
    Checks if an evaluation run name already exists for a given project.

    Args:
        eval_name (str): Name of the evaluation run
        project_name (str): Name of the project
        judgment_api_key (str): API key for authentication

    Raises:
        ValueError: If the evaluation run name already exists
        JudgmentAPIError: If there's an API error during the check
    """
    try:
        response = requests.post(
            f"{ROOT_API}/eval-run-name-exists/",
            json={
                "eval_name": eval_name,
                "project_name": project_name,
                "judgment_api_key": judgment_api_key,
            }
        )
        
        if response.status_code == 409:
            error(f"Evaluation run name '{eval_name}' already exists for this project")
            raise ValueError(f"Evaluation run name '{eval_name}' already exists for this project")
        
        if not response.ok:
            response_data = response.json()
            error_message = response_data.get('detail', 'An unknown error occurred.')
            error(f"Error checking eval run name: {error_message}")
            raise JudgmentAPIError(error_message)
            
    except requests.exceptions.RequestException as e:
        error(f"Failed to check if eval run name exists: {str(e)}")
        raise JudgmentAPIError(f"Failed to check if eval run name exists: {str(e)}")


def log_evaluation_results(merged_results: List[ScoringResult], evaluation_run: EvaluationRun) -> None:
    """
    Logs evaluation results to the Judgment API database.

    Args:
        merged_results (List[ScoringResult]): The results to log
        evaluation_run (EvaluationRun): The evaluation run containing project info and API key

    Raises:
        JudgmentAPIError: If there's an API error during logging
        ValueError: If there's a validation error with the results
    """
    try:
        res = requests.post(
            JUDGMENT_EVAL_LOG_API_URL,
            json={
                "results": [result.to_dict() for result in merged_results],
                "judgment_api_key": evaluation_run.judgment_api_key,
                "project_name": evaluation_run.project_name,
                "eval_name": evaluation_run.eval_name,
            }
        )
        
        if not res.ok:
            response_data = res.json()
            error_message = response_data.get('detail', 'An unknown error occurred.')
            error(f"Error {res.status_code}: {error_message}")
            raise JudgmentAPIError(error_message)
        
        if "ui_results_url" in res.json():
            rprint(f"\n🔍 You can view your evaluation results here: [rgb(106,0,255)]{res.json()['ui_results_url']}[/]\n")
            
    except requests.exceptions.RequestException as e:
        error(f"Request failed while saving evaluation results to DB: {str(e)}")
        raise JudgmentAPIError(f"Request failed while saving evaluation results to DB: {str(e)}")
    except Exception as e:
        error(f"Failed to save evaluation results to DB: {str(e)}")
        raise ValueError(f"Failed to save evaluation results to DB: {str(e)}")


def run_eval(evaluation_run: EvaluationRun, override: bool = False) -> List[ScoringResult]:
    """
    Executes an evaluation of `Example`s using one or more `Scorer`s

    Args:
        evaluation_run (EvaluationRun): Stores example and evaluation together for running
    
        Args: 
            project_name (str): The name of the project the evaluation results belong to
            eval_name (str): The name of the evaluation run
            examples (List[Example]): The examples to evaluate
            scorers (List[Union[JudgmentScorer, JudgevalScorer]]): A list of scorers to use for evaluation
            model (str): The model used as a judge when using LLM as a Judge
            aggregator (Optional[str]): The aggregator to use for evaluation if using Mixture of Judges
            metadata (Optional[Dict[str, Any]]): Additional metadata to include for this evaluation run, e.g. comments, dataset name, purpose, etc.
            judgment_api_key (Optional[str]): The API key for running evaluations on the Judgment API
            log_results (bool): Whether to log the results to the Judgment API


    Returns:
        List[ScoringResult]: The results of the evaluation. Each result is a dictionary containing the fields of a `ScoringResult` object.
    """
    
    # Call endpoint to check to see if eval run name exists (if we DON'T want to override and DO want to log results)
    if not override and evaluation_run.log_results:
        check_eval_run_name_exists(
            evaluation_run.eval_name,
            evaluation_run.project_name,
            evaluation_run.judgment_api_key
        )
    
    # Set example IDs if not already set
    debug("Initializing examples with IDs and timestamps")
    for idx, example in enumerate(evaluation_run.examples):
        if example.example_id is None:
            example.example_id = idx
            debug(f"Set example ID {idx} for input: {example.input[:50]}...")
        example.timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")
        with example_logging_context(example.timestamp, example.example_id):
            debug(f"Initialized example {example.example_id}")
            debug(f"Input: {example.input}")
            debug(f"Actual output: {example.actual_output}")
            if example.expected_output:
                debug(f"Expected output: {example.expected_output}")
            if example.context:
                debug(f"Context: {example.context}")
            if example.retrieval_context:
                debug(f"Retrieval context: {example.retrieval_context}")
    
    debug(f"Starting evaluation run with {len(evaluation_run.examples)} examples")
    
    # Group APIJudgmentScorers and JudgevalScorers, then evaluate them in parallel
    debug("Grouping scorers by type")
    judgment_scorers: List[APIJudgmentScorer] = []
    local_scorers: List[JudgevalScorer] = []
    for scorer in evaluation_run.scorers:
        if isinstance(scorer, (APIJudgmentScorer, ClassifierScorer)):
            judgment_scorers.append(scorer)
            debug(f"Added judgment scorer: {type(scorer).__name__}")
        else:
            local_scorers.append(scorer)
            debug(f"Added local scorer: {type(scorer).__name__}")
    
    debug(f"Found {len(judgment_scorers)} judgment scorers and {len(local_scorers)} local scorers")
    
    api_results: List[ScoringResult] = []
    local_results: List[ScoringResult] = []
    
    # Execute evaluation using Judgment API
    if judgment_scorers:
        info("Starting API evaluation")
        debug(f"Creating API evaluation run with {len(judgment_scorers)} scorers")
        try:  # execute an EvaluationRun with just JudgmentScorers
            api_evaluation_run: EvaluationRun = EvaluationRun(
                eval_name=evaluation_run.eval_name,
                project_name=evaluation_run.project_name,
                examples=evaluation_run.examples,
                scorers=judgment_scorers,
                model=evaluation_run.model,
                aggregator=evaluation_run.aggregator,
                metadata=evaluation_run.metadata,
                judgment_api_key=evaluation_run.judgment_api_key,
                log_results=evaluation_run.log_results
            )
            debug("Sending request to Judgment API")    
            response_data: List[Dict] = execute_api_eval(api_evaluation_run)  # Dicts are `ScoringResult` objs
            info(f"Received {len(response_data['results'])} results from API")
        except JudgmentAPIError as e:
            error(f"An error occurred while executing the Judgment API request: {str(e)}")
            raise JudgmentAPIError(f"An error occurred while executing the Judgment API request: {str(e)}")
        except ValueError as e:
            raise ValueError(f"Please check your EvaluationRun object, one or more fields are invalid: {str(e)}")
        
        # Convert the response data to `ScoringResult` objects
        debug("Processing API results")
        for idx, result in enumerate(response_data["results"]):  
            with example_logging_context(evaluation_run.examples[idx].timestamp, evaluation_run.examples[idx].example_id):
                for scorer in judgment_scorers:
                    debug(f"Processing API result for example {idx} and scorer {scorer.score_type}")
                # filter for key-value pairs that are used to initialize ScoringResult
                # there may be some stuff in here that doesn't belong in ScoringResult
                # TODO: come back and refactor this to have ScoringResult take in **kwargs
                filtered_result = {k: v for k, v in result.items() if k in ScoringResult.__annotations__}
                
                # Convert scorers_data dicts to ScorerData objects
                if "scorers_data" in filtered_result and filtered_result["scorers_data"]:
                    filtered_result["scorers_data"] = [
                        ScorerData(**scorer_dict) 
                        for scorer_dict in filtered_result["scorers_data"]
                    ]
                
                api_results.append(ScoringResult(**filtered_result))

    # Run local evals
    if local_scorers:  # List[JudgevalScorer]
        info("Starting local evaluation")
        for example in evaluation_run.examples:
            with example_logging_context(example.timestamp, example.example_id):
                debug(f"Processing example {example.example_id}: {example.input}")
        
        results: List[ScoringResult] = asyncio.run(
            a_execute_scoring(
                evaluation_run.examples,
                local_scorers,
                model=evaluation_run.model,
                ignore_errors=True,
                skip_on_missing_params=True,
                show_indicator=True,
                _use_bar_indicator=True,
                throttle_value=0,
                max_concurrent=100,
            )
        )
        local_results = results
        info(f"Local evaluation complete with {len(local_results)} results")

    # Aggregate the ScorerData from the API and local evaluations
    debug("Merging API and local results")
    merged_results: List[ScoringResult] = merge_results(api_results, local_results)
    merged_results = check_missing_scorer_data(merged_results)

    info(f"Successfully merged {len(merged_results)} results")

    if evaluation_run.log_results:
        log_evaluation_results(merged_results, evaluation_run)

    for i, result in enumerate(merged_results):
        if not result.scorers_data:  # none of the scorers could be executed on this example
            info(f"None of the scorers could be executed on example {i}. This is usually because the Example is missing the fields needed by the scorers. Try checking that the Example has the necessary fields for your scorers.")
    return merged_results

def assert_test(scoring_results: List[ScoringResult]) -> None:
    """
    Collects all failed scorers from the scoring results.

    Args:
        ScoringResults (List[ScoringResult]): List of scoring results to check

    Returns:
        None. Raises exceptions for any failed test cases.
    """
    failed_cases: List[ScorerData] = []

    for result in scoring_results:
        if not result.success:

            # Create a test case context with all relevant fields
            test_case = {
                'input': result.input,
                'actual_output': result.actual_output,
                'expected_output': result.expected_output,
                'context': result.context,
                'retrieval_context': result.retrieval_context,
                'eval_run_name': result.eval_run_name,
                'failed_scorers': []
            }
            if result.scorers_data:
                # If the result was not successful, check each scorer_data
                for scorer_data in result.scorers_data:
                    if not scorer_data.success:
                        test_case['failed_scorers'].append(scorer_data)
            failed_cases.append(test_case)

    if failed_cases:
        error_msg = f"The following test cases failed: \n"
        for fail_case in failed_cases:
            error_msg += f"\nInput: {fail_case['input']}\n"
            error_msg += f"Actual Output: {fail_case['actual_output']}\n"
            error_msg += f"Expected Output: {fail_case['expected_output']}\n"
            error_msg += f"Context: {fail_case['context']}\n"
            error_msg += f"Retrieval Context: {fail_case['retrieval_context']}\n"
            error_msg += f"Eval Run Name: {fail_case['eval_run_name']}\n"
    
            for fail_scorer in fail_case['failed_scorers']:

                error_msg += (
                    f"\nScorer Name: {fail_scorer.name}\n"
                    f"Threshold: {fail_scorer.threshold}\n"
                    f"Success: {fail_scorer.success}\n" 
                    f"Score: {fail_scorer.score}\n"
                    f"Reason: {fail_scorer.reason}\n"
                    f"Strict Mode: {fail_scorer.strict_mode}\n"
                    f"Evaluation Model: {fail_scorer.evaluation_model}\n"
                    f"Error: {fail_scorer.error}\n"
                    f"Evaluation Cost: {fail_scorer.evaluation_cost}\n"
                    f"Verbose Logs: {fail_scorer.verbose_logs}\n"
                    f"Additional Metadata: {fail_scorer.additional_metadata}\n"
                )
            error_msg += "-"*100
    
        raise AssertionError(error_msg)
    


"""
This script is a cookbook of how to create a custom scorer using a ClassifierScorer.

Simply use a natural language prompt and guide the LLM to output a score based on the input by 
choosing from a set of options.
"""

from judgeval import JudgmentClient
from judgeval.scorers import ClassifierScorer
from judgeval.data import Example

text2sql_scorer = ClassifierScorer(
    "Text to SQL",
    slug="text2sql-487126418",
    threshold=1.0,
    conversation=[{
        "role": "system",
        "content": """You will be given a natural language query, a corresponding LLM-generated SQL query, and a table schema + (optional) metadata.

** TASK INSTRUCTIONS **
Your task is to decide whether the LLM generated SQL query properly filters for what the natural language query is asking, based on the table schema + (optional) metadata. 
Additionally, you should check if the SQL query is valid based on the table schema (checking for syntax errors, false column names, etc.)

** TIPS **
- Look for correct references to the table schema for column names, table names, etc.
- Check that the SQL query can be executed; make sure JOINs, GROUP BYs, ORDER BYs, etc. are valid with respect to the table schema.
- Check that aggregation functions (COUNT, SUM, AVG, etc.) are used appropriately with GROUP BY clauses
- Verify that WHERE conditions use the correct operators and data types for comparisons
- Ensure LIMIT and OFFSET clauses make sense for the query's purpose
- Check that JOINs use the correct keys and maintain referential integrity
- Verify that ORDER BY clauses use valid column names and sort directions
- Check for proper handling of NULL values where relevant
- Ensure subqueries are properly constructed and correlated when needed
- EVEN IF THE QUERY IS VALID, IF IT DOESN'T WORK FOR THE NATURAL LANGUAGE QUERY, YOU SHOULD CHOOSE "N" AS THE ANSWER.

** FORMATTING YOUR ANSWER **
If the SQL query is valid and works for the natural language query, choose option "Y" and otherwise "N". Provide a justification for your decision; if you choose "N", explain what about the LLM-generated SQL query is incorrect, or explain why it doesn't address the natural language query. 
IF YOUR JUSTIFICATION SHOWS THAT THE SQL QUERY IS VALID AND WORKS FOR THE NATURAL LANGUAGE QUERY, YOU SHOULD CHOOSE "Y" AS THE ANSWER. 
IF THE SQL QUERY IS INVALID, YOU SHOULD CHOOSE "N" AS THE ANSWER. 

** YOUR TURN **
Natural language query:
{{input}}

LLM generated SQL query:
{{actual_output}}

Table schema:
{{context}}
        """
    }],
    options={
        "Y": 1.0, 
        "N": 0.0
    }
)


if __name__ == "__main__":
    client = JudgmentClient()

    table_schema = """CREATE TABLE Artists (
    artist_id VARCHAR(50) PRIMARY KEY,
    name VARCHAR(255) NOT NULL,
    genre VARCHAR(100),
    followers INT,
    popularity INT,
    created_at TIMESTAMP DEFAULT CURRENT_TIMESTAMP
);

CREATE TABLE Albums (
    album_id VARCHAR(50) PRIMARY KEY,
    title VARCHAR(255) NOT NULL,
    artist_id VARCHAR(50) NOT NULL,
    release_date DATE,
    total_tracks INT,
    album_type VARCHAR(50) CHECK (album_type IN ('album', 'single', 'compilation')),
    created_at TIMESTAMP DEFAULT CURRENT_TIMESTAMP,
    FOREIGN KEY (artist_id) REFERENCES Artists(artist_id) ON DELETE CASCADE
);

CREATE TABLE Tracks (
    track_id VARCHAR(50) PRIMARY KEY,
    title VARCHAR(255) NOT NULL,
    album_id VARCHAR(50) NOT NULL,
    artist_id VARCHAR(50) NOT NULL,
    duration_ms INT NOT NULL,
    explicit BOOLEAN DEFAULT FALSE,
    popularity INT DEFAULT 0,
    preview_url VARCHAR(255),
    created_at TIMESTAMP DEFAULT CURRENT_TIMESTAMP,
    FOREIGN KEY (album_id) REFERENCES Albums(album_id) ON DELETE CASCADE,
    FOREIGN KEY (artist_id) REFERENCES Artists(artist_id) ON DELETE CASCADE
);

CREATE TABLE Users (
    user_id VARCHAR(50) PRIMARY KEY,
    username VARCHAR(100) NOT NULL UNIQUE,
    email VARCHAR(255) NOT NULL UNIQUE,
    created_at TIMESTAMP DEFAULT CURRENT_TIMESTAMP
);

CREATE TABLE Playlists (
    playlist_id VARCHAR(50) PRIMARY KEY,
    user_id VARCHAR(50) NOT NULL,
    name VARCHAR(255) NOT NULL,
    description TEXT,
    created_at TIMESTAMP DEFAULT CURRENT_TIMESTAMP,
    FOREIGN KEY (user_id) REFERENCES Users(user_id) ON DELETE CASCADE
);

CREATE TABLE PlaylistTracks (
    playlist_id VARCHAR(50),
    track_id VARCHAR(50),
    added_at TIMESTAMP DEFAULT CURRENT_TIMESTAMP,
    PRIMARY KEY (playlist_id, track_id),
    FOREIGN KEY (playlist_id) REFERENCES Playlists(playlist_id) ON DELETE CASCADE,
    FOREIGN KEY (track_id) REFERENCES Tracks(track_id) ON DELETE CASCADE
);

CREATE TABLE UserListeningHistory (
    history_id SERIAL PRIMARY KEY,
    user_id VARCHAR(50) NOT NULL,
    track_id VARCHAR(50) NOT NULL,
    listened_at TIMESTAMP DEFAULT CURRENT_TIMESTAMP,
    FOREIGN KEY (user_id) REFERENCES Users(user_id) ON DELETE CASCADE,
    FOREIGN KEY (track_id) REFERENCES Tracks(track_id) ON DELETE CASCADE
);
"""

    all_tracks_one_artist_correct = Example(
        input="Find all tracks by the artist 'Drake', sorted by popularity.",
        actual_output="""SELECT t.track_id, t.title, t.popularity, a.name AS artist_name
FROM Tracks t
JOIN Artists a ON t.artist_id = a.artist_id
WHERE a.name = 'Drake'
ORDER BY t.popularity DESC;
""",
        retrieval_context=[table_schema]
    )

    most_listened_to_one_user_correct = Example(
        input="Find the most listened to track by user 'user123'.",
        actual_output="""SELECT t.track_id, t.title, COUNT(uh.history_id) AS play_count
FROM UserListeningHistory uh
JOIN Tracks t ON uh.track_id = t.track_id
WHERE uh.user_id = 'user123'
GROUP BY t.track_id, t.title
ORDER BY play_count DESC
LIMIT 1;
""",
        retrieval_context=[table_schema]
    )

    highest_num_playlists_correct = Example(
        input="Find the 5 users with the highest number of playlists.",
        actual_output="""SELECT u.user_id, u.username, COUNT(p.playlist_id) AS total_playlists
FROM Users u
JOIN Playlists p ON u.user_id = p.user_id
GROUP BY u.user_id, u.username
ORDER BY total_playlists DESC
LIMIT 5;
""",
        retrieval_context=[table_schema]
    )

    most_popular_tracks_all_users_correct = Example(
        input="Find the 10 most popular tracks across all users.",
        actual_output="""SELECT t.track_id, t.title, COUNT(uh.history_id) AS total_listens
FROM Tracks t
JOIN UserListeningHistory uh ON t.track_id = uh.track_id
GROUP BY t.track_id, t.title
ORDER BY total_listens DESC
LIMIT 10;
""",
        retrieval_context=[table_schema]
    )

    most_popular_tracks_all_users_incorrect = Example(
        input="Find the 10 most popular tracks across all users.",
        actual_output="""SELECT t.track_user, t.title, COUNT(uh.history_id) AS total_listens
FROM Tracks t
JOIN UserHistory uh ON t.track_user = uh.track_user
GROUP BY t.track_user, t.title
ORDER BY total_listens DESC
LIMIT 10;
""",
        retrieval_context=[table_schema]
    )


    client.run_evaluation(
        examples=[
            all_tracks_one_artist_correct, 
            most_listened_to_one_user_correct, 
            highest_num_playlists_correct, 
            most_popular_tracks_all_users_correct, 
            most_popular_tracks_all_users_incorrect
        ],
        scorers=[text2sql_scorer],
        model="gpt-4o-mini",
        project_name="text2sql",
        eval_run_name="text2sql_test",
        override=True
    )



"""
Classes for representing examples in a dataset.
"""


from typing import TypeVar, Optional, Any, Dict, List
from pydantic import BaseModel
from enum import Enum
from datetime import datetime


Input = TypeVar('Input')
Output = TypeVar('Output')

class ExampleParams(Enum):
    INPUT = "input"
    ACTUAL_OUTPUT = "actual_output"
    EXPECTED_OUTPUT = "expected_output"
    CONTEXT = "context"
    RETRIEVAL_CONTEXT = "retrieval_context"
    TOOLS_CALLED = "tools_called"
    EXPECTED_TOOLS = "expected_tools"
    REASONING = "reasoning"


class Example(BaseModel):
    input: Input
    actual_output: Output
    expected_output: Optional[str] = None
    context: Optional[List[str]] = None
    retrieval_context: Optional[List[str]] = None
    additional_metadata: Optional[Dict[str, Any]] = None
    tools_called: Optional[List[str]] = None
    expected_tools: Optional[List[str]] = None
    name: Optional[str] = None
    example_id: Optional[str] = None
    timestamp: Optional[str] = None
    trace_id: Optional[str] = None

    def __init__(self, **data):
        super().__init__(**data)
        # Set timestamp if not provided
        if self.timestamp is None:
            self.timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")

    def to_dict(self):
        return {
            "input": self.input,
            "actual_output": self.actual_output,
            "expected_output": self.expected_output,
            "context": self.context,
            "retrieval_context": self.retrieval_context,
            "additional_metadata": self.additional_metadata,
            "tools_called": self.tools_called,
            "expected_tools": self.expected_tools,
            "name": self.name,
            "example_id": self.example_id,
            "timestamp": self.timestamp,
            "trace_id": self.trace_id
        }

    def __str__(self):
        return (
            f"Example(input={self.input}, "
            f"actual_output={self.actual_output}, "
            f"expected_output={self.expected_output}, "
            f"context={self.context}, "
            f"retrieval_context={self.retrieval_context}, "
            f"additional_metadata={self.additional_metadata}, "
            f"tools_called={self.tools_called}, "
            f"expected_tools={self.expected_tools}, "
            f"name={self.name}, "
            f"example_id={self.example_id}, "
            f"timestamp={self.timestamp}, "
            f"trace_id={self.trace_id})"
        )



"""
Code for the local implementation of the Faithfulness metric.
"""
from typing import List, Optional, Union

from judgeval.constants import APIScorer
from judgeval.data import (
    Example, 
    ExampleParams
)
from judgeval.scorers import JudgevalScorer
from judgeval.scorers.utils import (
    get_or_create_event_loop, 
    check_example_params
)
from judgeval.judges.utils import create_judge
from judgeval.judges import JudgevalJudge
from judgeval.scorers.utils import (
    scorer_progress_meter, 
    create_verbose_logs, 
    parse_response_json
)
from judgeval.scorers.judgeval_scorers.local_implementations.faithfulness.prompts import (
    FaithfulnessTemplate,
    FaithfulnessVerdict,
    Verdicts,
    Reason,
    Claims,
)


required_params = [
    ExampleParams.INPUT,
    ExampleParams.ACTUAL_OUTPUT,
    ExampleParams.RETRIEVAL_CONTEXT,
]


class FaithfulnessScorer(JudgevalScorer):
    def __init__(
        self,
        threshold: float = 0.5,
        model: Optional[Union[str, JudgevalJudge]] = None,
        include_reason: bool = True,
        async_mode: bool = True,
        strict_mode: bool = False,
        verbose_mode: bool = False,
        user: Optional[str] = None
    ):
        super().__init__(
            score_type=APIScorer.FAITHFULNESS,
            threshold=1 if strict_mode else threshold,
            evaluation_model=None,
            include_reason=include_reason,
            async_mode=async_mode,
            strict_mode=strict_mode,
            verbose_mode=verbose_mode
        )
        self.user = user
        self.model, self.using_native_model = create_judge(model)
        self.using_native_model = True  # NOTE: SETTING THIS FOR LITELLM and TOGETHER usage
        self.evaluation_model = self.model.get_model_name()

    def score_example(
        self,
        example: Example,
        all_claims: bool = False,
        _show_indicator: bool = True,
    ) -> float:
        check_example_params(example, required_params, self)
        
        with scorer_progress_meter(self, display_meter=_show_indicator):
            if self.async_mode:
                loop = get_or_create_event_loop()
                loop.run_until_complete(
                    self.a_score_example(
                        example, 
                        all_claims=all_claims, 
                        _show_indicator=False
                    )
                )
            else:
                self.claims = self._generate_claims(example.actual_output, all_claims=all_claims)
                if self.additional_metadata is None:
                    self.additional_metadata = {}
                self.additional_metadata["claims"] = self.claims  # Add claims generated to metadata

                self.verdicts = self._generate_verdicts(example.retrieval_context)
                self.additional_metadata["verdicts"] = [v.model_dump() for v in self.verdicts]  # Add verdicts generated to metadata
                
                self.score = self._calculate_score()
                self.reason = self._generate_reason()
                self.success = self.score >= self.threshold
                self.verbose_logs = create_verbose_logs(
                    self,
                    steps=[
                        f"Claims:\n{self.claims}",
                        f"Verdicts:\n{self.verdicts}",
                        f"Score: {self.score}\nReason: {self.reason}",
                    ],
                )

                return self.score

    async def a_score_example(
        self,
        example: Example,
        _show_indicator: bool = True
    ) -> float:
        check_example_params(example, required_params, self)

        with scorer_progress_meter(
            self, async_mode=True, display_meter=_show_indicator
        ):
            self.claims = await self._a_generate_claims(example.actual_output)

            if self.additional_metadata is None:
                self.additional_metadata = {}
            self.additional_metadata["claims"] = self.claims

            self.verdicts = await self._a_generate_verdicts(example.retrieval_context)  
            self.additional_metadata["verdicts"] = [v.model_dump() for v in self.verdicts]  # Add verdicts generated to metadata

            self.score = self._calculate_score()
            self.reason = await self._a_generate_reason()
            self.success = self.score >= self.threshold
            self.verbose_logs = create_verbose_logs(
                self,
                steps=[
                    f"Claims:\n{self.claims}",
                    f"Verdicts:\n{self.verdicts}",
                    f"Score: {self.score}\nReason: {self.reason}",
                ],
            )

            return self.score

    async def _a_generate_reason(self) -> str:
        if self.include_reason is False:
            return None

        contradictions = []
        for verdict in self.verdicts:
            if verdict.verdict.strip().lower() == "no":
                contradictions.append(verdict.model_dump())

        prompt: dict = FaithfulnessTemplate.justify_reason(
            contradictions=contradictions,
            score=format(self.score, ".2f"),
        )
        if self.using_native_model:
            res = await self.model.a_generate(prompt)
            data = parse_response_json(res, self)
            return data["reason"]
        else:
            try:
                res: Reason = await self.model.a_generate(prompt, schema=Reason)
                return res.reason
            except TypeError:
                res = await self.model.a_generate(prompt)
                data = parse_response_json(res, self)
                return data["reason"]

    def _generate_reason(self) -> str:
        if self.include_reason is False:
            return None

        contradictions = []
        for verdict in self.verdicts:
            if verdict.verdict.strip().lower() == "no":
                contradictions.append(verdict.reason)

        prompt: dict = FaithfulnessTemplate.justify_reason(
            contradictions=contradictions,
            score=format(self.score, ".2f"),
        )

        if self.using_native_model:
            res = self.model.generate(prompt)
            data = parse_response_json(res, self)
            return data["reason"]
        else:
            try:
                res: Reason = self.model.generate(prompt, schema=Reason)
                return res.reason
            except TypeError:
                res = self.model.generate(prompt)
                data = parse_response_json(res, self)
                return data["reason"]

    async def _a_generate_verdicts(self, retrieval_context: str) -> List[FaithfulnessVerdict]:
        if len(self.claims) == 0:
            return []

        verdicts: List[FaithfulnessVerdict] = []

        claims = [
            claim["claim"] for claim in self.claims
        ]  # We only need the claims, not the quotes involved

        prompt = FaithfulnessTemplate.create_verdicts(
            claims=claims,
            retrieval_context=retrieval_context,
        )
        if self.using_native_model:
            res = await self.model.a_generate(prompt)
            data = parse_response_json(res, self)
            verdicts = [
                FaithfulnessVerdict(**item) for item in data["verdicts"]
            ]
            return verdicts
        else:
            try:
                res: Verdicts = await self.model.generate(
                    prompt, schema=Verdicts
                )
                verdicts = [item for item in res.verdicts]
                return verdicts
            except TypeError:
                res = await self.model.a_generate(prompt)
                data = parse_response_json(res, self)
                verdicts = [
                    FaithfulnessVerdict(**item) for item in data["verdicts"]
                ]
                return verdicts

    def _generate_verdicts(self, retrieval_context: str) -> List[FaithfulnessVerdict]:
        if len(self.claims) == 0:
            return []

        verdicts: List[FaithfulnessVerdict] = []

        claims = [
            claim["claim"] for claim in self.claims
        ]  # We only need the claims, not the quotes involved

        prompt = FaithfulnessTemplate.create_verdicts(
            claims=claims,
            retrieval_context=retrieval_context,
        )
        if self.using_native_model:
            res = self.model.generate(prompt)
            data = parse_response_json(res, self)
            verdicts = [
                FaithfulnessVerdict(**item) for item in data["verdicts"]
            ]
            return verdicts
        else:
            try:
                res: Verdicts = self.model.generate(prompt, schema=Verdicts)
                verdicts = [item for item in res.verdicts]
                return verdicts
            except TypeError:
                res = self.model.generate(prompt)
                data = parse_response_json(res, self)
                verdicts = [
                    FaithfulnessVerdict(**item) for item in data["verdicts"]
                ]
                return verdicts

    async def _a_generate_claims(self, actual_output: str) -> List[str]:
        prompt = FaithfulnessTemplate.find_claims(text=actual_output)
        if self.using_native_model:
            res = await self.model.a_generate(prompt)
            data = parse_response_json(res, self)
            return data["claims"]
        else:
            try:
                res: Claims = await self.model.a_generate(prompt, schema=Claims)
                return res.claims
            except TypeError:
                res = await self.model.a_generate(prompt)
                data = parse_response_json(res, self)
                return data["claims"]

    def _generate_claims(self, actual_output: str, all_claims: bool = False) -> List[str]:
        prompt = FaithfulnessTemplate.find_claims(text=actual_output)
        if self.using_native_model:
            res = self.model.generate(prompt)
            data = parse_response_json(res, self)
            return data["claims"]
        else:
            try:
                res: Claims = self.model.generate(prompt, schema=Claims)
                return res.claims
            except TypeError:
                res = self.model.generate(prompt)
                data = parse_response_json(res, self)
                return data["claims"]

    def _calculate_score(self) -> float:
        number_of_verdicts = len(self.verdicts)
        if number_of_verdicts == 0:
            return 1

        faithfulness_count = 0
        for verdict in self.verdicts:
            if verdict.verdict.strip().lower() != "no":
                faithfulness_count += 1

        score = faithfulness_count / number_of_verdicts
        return 0 if self.strict_mode and score < self.threshold else score

    def _success_check(self) -> bool:
        if self.error is not None:
            self.success = False
        else:
            try:
                self.success = self.score >= self.threshold
            except:
                self.success = False
        return self.success

    def get_claims(self):
        return self.claims
    
    def get_verdicts(self):
        return self.verdicts

    @property
    def __name__(self):
        return "Faithfulness"
    


"""
Util prompts for AnswerCorrectnessScorer
"""

from typing import List, Optional, Tuple
from pydantic import BaseModel, Field


# BaseModels to enforce formatting in LLM JSON response
class Statements(BaseModel):
    statements: List[str]


class ACVerdict(BaseModel):
    verdict: str
    reason: str


class Verdicts(BaseModel):
    verdicts: List[ACVerdict]


class Reason(BaseModel):
    reason: str


class AnswerCorrectnessTemplate:
    @staticmethod
    def deduce_statements(expected_output):
        return f"""You will be presented with a piece of text. Your task is to break down the text and generate a list of statements contained within the text. Single words and ambiguous phrases should be considered statements.

===== START OF EXAMPLES =====
Example 1:
Example text: The weather is sunny today. Temperature is 75 degrees. Don't forget your sunscreen!

Output:
{{
    "statements": ["The weather is sunny today", "Temperature is 75 degrees", "Don't forget your sunscreen!"]
}}

Example 2:
Example text: I love pizza. It has cheese and tomato sauce and the crust is crispy.

Output:
{{
    "statements": ["I love pizza", "It has cheese and tomato sauce", "The crust is crispy"]
}}
===== END OF EXAMPLES =====

        
**
IMPORTANT: Please return your answer in valid JSON format, with the "statements" key mapping to a list of strings. No words or explanation is needed.
**

==== START OF INPUT ====
Text:
{expected_output}
==== END OF INPUT ====

==== YOUR ANSWER ====
JSON:
"""

    @staticmethod
    def generate_verdicts(statements, actual_output):
        return f"""You will be provided with:\n
- a list of statements from a text that we will refer to as expected output
- a text that we will refer to as actual output\n

Your task is to determine whether each statement from the expected output is correct/consistent with the actual output text.
More specifically, you should generate a JSON object with the key "verdicts". "verdicts" will map to a list of nested JSON objects with two keys: `verdict` and `reason`.
The "reason" key should provide an explanation for your choice, regardless of which verdict you select. Try providing quotes from the text(s) to justify your answer where possible.
The "verdict" key be EXACTLY EITHER "yes" or "no". You should select "yes" if the statement is correct/consistent based on the actual output and "no" otherwise.

==== OUTPUT FORMATTING ====
IMPORTANT: Please make sure to only return in JSON format, with the "verdicts" key mapping to a list of JSON objects. Each JSON object should contain keys "verdict" (one of "yes" or "no") and "reason" (str). 

==== START OF EXAMPLES ====
Example input 1: What's the capital of France?
Example expected output statements 1: ["Paris is the capital of France", "It is located in northern France", "The city has a population of over 2 million"]
Example actual output 1: "Paris is the capital city of France. It is situated in the northern part of the country and has over 2 million residents."
Example JSON 1:
{{
    "verdicts": [
        {{
            "verdict": "yes",
            "reason": "The actual output directly states 'Paris is the capital city of France', which matches the statement"
        }},
        {{
            "verdict": "yes", 
            "reason": "The actual output confirms this by saying it is 'situated in the northern part of the country'"
        }},
        {{
            "verdict": "yes",
            "reason": "The actual output mentions the city 'has over 2 million residents', matching the population statement"
        }}
    ]
}}

Example input 2: What is the largest planet in our solar system?
Example expected output statements 2: ["Jupiter is the largest planet", "It is a gas giant", "Jupiter has 79 known moons", "The Great Red Spot is a storm on Jupiter"]
Example actual output 2: "Jupiter is the biggest planet in the solar system. It is made mostly of gas. The planet has many moons orbiting it."
Example JSON 2:
{{
    "verdicts": [
        {{
            "verdict": "yes",
            "reason": "The actual output confirms 'Jupiter is the biggest planet', which is equivalent to it being the largest"
        }},
        {{
            "verdict": "yes",
            "reason": "The actual output states it is 'made mostly of gas', indicating it is a gas giant"
        }},
        {{
            "verdict": "no",
            "reason": "While the actual output mentions Jupiter has 'many moons', it does not specify the exact number of 79 known moons"
        }},
        {{
            "verdict": "no",
            "reason": "The actual output makes no mention of the Great Red Spot or any storms on Jupiter"
        }}
    ]
}}
==== END OF EXAMPLES ====

** LASTLY **
Since you are tasked to choose a verdict for each statement, the number of "verdicts" SHOULD BE EXACTLY EQUAL to the number of "statements".


==== YOUR TURN =====

Statements:
{statements}

Actual output:
{actual_output}

JSON:
"""

    @staticmethod
    def generate_reason(incorrect_statements: List[Tuple[str, str]], score: float):
        incorrect_statements = "\n".join([f"statement: {statement}\nreason: {reason}\n------" for statement, reason in incorrect_statements])
        return f"""==== TASK INSTRUCTIONS ====\nYou will provided with two inputs: an answer correctness score and a list of inconsistent/incorrect statements made in a model's output (with the reason why it's irrelevant). Your task is to provide a CLEAR and CONCISE reason for the answer correctness score. 
For context, there were a list of statements generated from an expected output. The model's actual output was then compared to the expected output, and we collected a list of claims made in the expected output that were either incorrect or inconsistent with the actual output. 
The score represents how well the model's output matches the expected output.
You should explain why the score is not higher, but also include why its current score is fair.
The incorrect statements represent parts of the model output that are incorrect or inconsistent with the expected output. The incorrect statement will be paired with the reason why it's incorrect.
If there are no incorrect statements, instead respond with a positive remark with an upbeat encouraging tone (but don't overblow the kind attitude).


==== FORMATTING YOUR ANSWER ====
IMPORTANT: Please make sure to only return in JSON format, with the 'reason' key providing the reason.
Example JSON:
{{
    "reason": "The score is <answer_relevancy_score> because <your_reason>."
}}

==== YOUR TURN ====
---- ANSWER CORRECTNESS SCORE ----
{score}

---- INCORRECT STATEMENTS ----
{incorrect_statements}

---- YOUR RESPONSE ----
JSON:
"""




"""
Constant variables used throughout source code
"""

from enum import Enum
import litellm
import os

class APIScorer(str, Enum):  
    """
    Collection of proprietary scorers implemented by Judgment.

    These are ready-made evaluation scorers that can be used to evaluate
    Examples via the Judgment API.
    """
    FAITHFULNESS = "faithfulness"
    ANSWER_RELEVANCY = "answer_relevancy"
    ANSWER_CORRECTNESS = "answer_correctness"
    HALLUCINATION = "hallucination"
    SUMMARIZATION = "summarization"
    CONTEXTUAL_RECALL = "contextual_recall"
    CONTEXTUAL_RELEVANCY = "contextual_relevancy"
    CONTEXTUAL_PRECISION = "contextual_precision"
    TOOL_CORRECTNESS = "tool_correctness"
    JSON_CORRECTNESS = "json_correctness"

    @classmethod
    def _missing_(cls, value):
        # Handle case-insensitive lookup
        for member in cls:
            if member.value == value.lower():
                return member

ROOT_API = os.getenv("JUDGMENT_API_URL", "https://api.judgmentlabs.ai")
## API URLs
JUDGMENT_EVAL_API_URL = f"{ROOT_API}/evaluate/"
JUDGMENT_DATASETS_PUSH_API_URL = f"{ROOT_API}/datasets/push/"
JUDGMENT_DATASETS_PULL_API_URL = f"{ROOT_API}/datasets/pull/"
JUDGMENT_EVAL_LOG_API_URL = f"{ROOT_API}/log_eval_results/"
JUDGMENT_EVAL_FETCH_API_URL = f"{ROOT_API}/fetch_eval_results/"
JUDGMENT_TRACES_SAVE_API_URL = f"{ROOT_API}/traces/save/"

## Models
TOGETHER_SUPPORTED_MODELS = {
    "QWEN": "Qwen/Qwen2-72B-Instruct",
    "LLAMA3_70B_INSTRUCT_TURBO": "meta-llama/Meta-Llama-3.1-70B-Instruct-Turbo",
    "LLAMA3_405B_INSTRUCT_TURBO": "meta-llama/Meta-Llama-3.1-405B-Instruct-Turbo",
    "LLAMA3_8B_INSTRUCT_TURBO": "meta-llama/Llama-3.2-3B-Instruct-Turbo",
    "MISTRAL_8x22B_INSTRUCT": "mistralai/Mixtral-8x22B-Instruct-v0.1",
    "MISTRAL_8x7B_INSTRUCT": "mistralai/Mixtral-8x7B-Instruct-v0.1",
}

ACCEPTABLE_MODELS = set(litellm.model_list) | set(TOGETHER_SUPPORTED_MODELS.keys())

## System settings
MAX_WORKER_THREADS = 10



"""
This script implements a custom scorer to evaluate customer support responses.

It checks if a support response mentions competitors (like Adidas, Reebok, etc.) in a positive way.
The scorer penalizes responses that promote competitor products, helping maintain brand focus in
customer interactions. This would be useful to score customer support responses for Nike, for example.
"""


from judgeval import JudgmentClient
from judgeval.scorers import ClassifierScorer
from judgeval.data import Example


competitor_mentions_scorer = ClassifierScorer(
    "Competitor Mentions",
    slug="competitor_mentions-487126418",
    threshold=1.0,
    conversation=[{
        "role": "system",
        "content": """Does the following customer support response discuss any of the following competitors in a positive way? (Y/N). 
        
        Competitors: Adidas, Reebok, Hoka, ON, Converse
        
        Customer Question: {{input}}
        Customer Support Response: {{actual_output}}
        """
    }],
    options={
        "Y": 0.0, 
        "N": 1.0
    }
)


if __name__ == "__main__":
    client = JudgmentClient()

    positive_example = Example(
        input="What are the best shoes for running priced under $130?",
        actual_output="You'd want to check out the newest Nike Vaporfly, it's only $120 and built for performance. "
    )

    negative_example = Example(
        input="What are the best shoes for running priced under $130?",
        actual_output="The Nike Vaporfly is a great shoe built for performance. Other great options include the Adidas Ultraboost and the Reebok Nano X which are affordable and speedy."
    )

    client.run_evaluation(
        examples=[positive_example, negative_example],
        scorers=[competitor_mentions_scorer],
        model="gpt-4o-mini",
        project_name="competitor_mentions",
        eval_run_name="competitor_mentions_test",
    )






"""
Cookbook for CI testing LLM applications using `judgeval`

Includes unit tests and end-to-end tests for an OpenAI API-based travel agent
"""

import asyncio
import os
import pytest
from demo.cookbooks.openai_travel_agent.agent import *
from judgeval import JudgmentClient
from judgeval.data import Example
from judgeval.scorers import (
    AnswerCorrectnessScorer, 
    AnswerRelevancyScorer, 
    FaithfulnessScorer
)   


@pytest.fixture
def judgment_client():
    return JudgmentClient()


@pytest.fixture
def research_data():
    return {
        "attractions": [
            "The iconic Eiffel Tower stands at 324 meters tall and welcomes over 7 million visitors annually. Visitors can access three levels, with the top floor offering panoramic views of Paris. The tower features two restaurants: 58 Tour Eiffel and the Michelin-starred Le Jules Verne.",
            "The Louvre Museum houses over 380,000 objects and displays 35,000 works of art across eight departments. Home to the Mona Lisa and Venus de Milo, it's the world's largest art museum with 72,735 square meters of exhibition space. Visitors typically need 3-4 hours to see the highlights.",
            "The historic district of Montmartre sits on Paris's highest hill at 130 meters. Famous for the white Sacré-Cœur Basilica and Place du Tertre filled with artists, it was once home to renowned painters like Picasso and Van Gogh. The area retains its village-like charm with winding cobblestone streets and authentic Parisian cafes."
        ],
        "hotels": [
            "Hotel de la Paix is a luxurious 5-star establishment in the 16th arrondissement, featuring 85 rooms and suites decorated in classic Parisian style. The hotel offers a Michelin-starred restaurant, spa facilities, and is located just 10 minutes from the Arc de Triomphe.",
            "Hotel de Paris, situated in the Opera district, combines Belle Époque architecture with modern amenities. Recently renovated in 2022, it offers 107 rooms, a rooftop bar with Eiffel Tower views, and has received the Palace distinction for exceptional service.",
            "Hotel de Ville, a boutique hotel in Le Marais, occupies a restored 17th-century mansion. With 40 individually designed rooms, a courtyard garden, and acclaimed restaurant, it provides an authentic Parisian experience steps from Notre-Dame Cathedral."
        ],
        "flights": [
            "Multiple daily direct flights to Paris Charles de Gaulle (CDG) from major US cities. Air France and United Airlines operate regular routes from JFK, LAX, and Chicago O'Hare. Flight times range from 7-11 hours depending on departure city.",
            "From San Francisco International Airport (SFO), Air France operates a daily direct flight AF085 departing at 3:30 PM, arriving at CDG at 11:15 AM next day. United Airlines also offers UA990 with a similar schedule. Average flight time is 10 hours 45 minutes."
        ],
        "weather": "Paris in mid-February typically experiences cool winter conditions with average daytime temperatures ranging from 8-12°C (46-54°F). Current forecast shows mostly sunny conditions with occasional cloud cover. Morning temperatures around 6°C (43°F) rising to 12°C (54°F) by afternoon. Light breeze of 8-12 km/h expected with 20% chance of precipitation. Evening temperatures dropping to 4°C (39°F). UV index moderate at 3.",
        "vector_db_results": []
    }


@pytest.fixture
def sample_itinerary() -> str:
    """
    Loads the sample itinerary from the saved file
    """
    PATH_TO_ITINERARY = os.path.join(os.path.dirname(__file__), "travel_response.txt")
    with open(PATH_TO_ITINERARY, 'r') as file:
        return file.read()


@pytest.fixture
def expected_itinerary():
    return """5-Day Paris Itinerary (February 11-15, 2025)

Accommodation: Hotel de Paris in the Opera district
- Selected for its central location, rooftop bar with Eiffel Tower views, and recent 2022 renovation

Transportation: 
- Arrival via Air France flight AF085/United Airlines UA990 from SFO, landing at CDG at 11:15 AM

Weather Considerations:
- Pack warm clothing for temperatures between 4-12°C (39-54°F)
- Morning activities planned indoors due to cooler temperatures
- Outdoor activities scheduled during peak afternoon warmth

Day 1 (Feb 11):
- 11:15 AM: Arrival at CDG, transfer to Hotel de Paris
- 2:00 PM: Hotel check-in and refresh
- 3:30 PM: Visit the Eiffel Tower (taking advantage of afternoon warmth)
- 7:00 PM: Dinner at Le Jules Verne in the Eiffel Tower

Day 2 (Feb 12):
- 9:00 AM: Breakfast at hotel
- 10:00 AM: Louvre Museum visit (3-4 hours, indoor activity during cool morning)
- 2:30 PM: Late lunch in Opera district
- 4:00 PM: Rooftop bar at hotel for sunset views
- Evening: Dinner at hotel's restaurant

Day 3 (Feb 13):
- 10:00 AM: Visit Montmartre (during warming temperatures)
- 11:00 AM: Explore Sacré-Cœur Basilica
- 12:30 PM: Lunch at local cafe in Montmartre
- 2:00 PM: Artist square at Place du Tertre
- Evening: Dinner at authentic Parisian bistro

Day 4 (Feb 14):
- Morning: Arc de Triomphe visit (10-minute walk from hotel)
- Afternoon: Shopping and exploring Opera district
- Evening: Valentine's Day dinner at hotel's Michelin-starred restaurant

Day 5 (Feb 15):
- Morning: Leisurely breakfast
- Late morning: Check-out and departure

Note: Indoor alternatives planned in case of precipitation (20% chance). Schedule optimized around temperature peaks of 12°C in afternoons."""


def test_websearch_tool_answer_relevancy(judgment_client):
    query = "What is the weather like in San Francisco on February 11th, 2025?"
    results = search_tavily(query)

    example = Example(
        input=query,
        actual_output=str(results)
    )

    scorer = AnswerRelevancyScorer(threshold=0.8)
    
    judgment_client.assert_test(
        examples=[example],
        scorers=[scorer],
        model="gpt-4o-mini",
        project_name="travel_agent_tests",
        eval_run_name="websearch_relevancy_test",
        override=True
    )


def test_travel_planning_faithfulness(judgment_client, sample_itinerary, research_data):
    
    destination = "Paris, France"
    start_date = "February 11th, 2025"
    end_date = "February 15th, 2025"

    hotels_example = Example(
        input=f"Create a structured travel itinerary for a trip to {destination} from {start_date} to {end_date}.",
        actual_output=sample_itinerary,
        retrieval_context=research_data["hotels"]
    )

    flights_example = Example(
        input=f"Create a structured travel itinerary for a trip to {destination} from {start_date} to {end_date}.",
        actual_output=sample_itinerary,
        retrieval_context=research_data["flights"]
    )

    judgment_client.assert_test(
        examples=[hotels_example, flights_example],
        scorers=[FaithfulnessScorer(threshold=1.0)],
        model="gpt-4o",
        project_name="travel_agent_tests",
        eval_run_name="travel_planning_faithfulness_test",
        override=True
    )


def test_travel_planning_answer_correctness(judgment_client, sample_itinerary, expected_itinerary):
    
    destination = "Paris, France"
    start_date = "February 11th, 2025"
    end_date = "February 15th, 2025"

    example = Example(
        input=f"Create a structured travel itinerary for a trip to {destination} from {start_date} to {end_date}.",
        actual_output=sample_itinerary,
        expected_output=expected_itinerary
    )
    with pytest.raises(AssertionError):
        judgment_client.assert_test(
            examples=[example],
            scorers=[AnswerCorrectnessScorer(threshold=0.75)],
            model="gpt-4o",
            project_name="travel_agent_tests",
            eval_run_name="travel_planning_correctness_test",
            override=True
        )


def save_travel_response(destination, start_date, end_date, research_data, file_path):
    response = asyncio.run(create_travel_plan(destination, start_date, end_date, research_data))
    with open(file_path, 'w') as f:
        f.write(response)


if __name__ == "__main__":
    sample_research_data = {
        "attractions": [
            "The iconic Eiffel Tower stands at 324 meters tall and welcomes over 7 million visitors annually. Visitors can access three levels, with the top floor offering panoramic views of Paris. The tower features two restaurants: 58 Tour Eiffel and the Michelin-starred Le Jules Verne.",
            "The Louvre Museum houses over 380,000 objects and displays 35,000 works of art across eight departments. Home to the Mona Lisa and Venus de Milo, it's the world's largest art museum with 72,735 square meters of exhibition space. Visitors typically need 3-4 hours to see the highlights.",
            "The historic district of Montmartre sits on Paris's highest hill at 130 meters. Famous for the white Sacré-Cœur Basilica and Place du Tertre filled with artists, it was once home to renowned painters like Picasso and Van Gogh. The area retains its village-like charm with winding cobblestone streets and authentic Parisian cafes."
        ],
        "hotels": [
            "Hotel de la Paix is a luxurious 5-star establishment in the 16th arrondissement, featuring 85 rooms and suites decorated in classic Parisian style. The hotel offers a Michelin-starred restaurant, spa facilities, and is located just 10 minutes from the Arc de Triomphe.",
            "Hotel de Paris, situated in the Opera district, combines Belle Époque architecture with modern amenities. Recently renovated in 2022, it offers 107 rooms, a rooftop bar with Eiffel Tower views, and has received the Palace distinction for exceptional service.",
            "Hotel de Ville, a boutique hotel in Le Marais, occupies a restored 17th-century mansion. With 40 individually designed rooms, a courtyard garden, and acclaimed restaurant, it provides an authentic Parisian experience steps from Notre-Dame Cathedral."
        ],
        "flights": [
            "Multiple daily direct flights to Paris Charles de Gaulle (CDG) from major US cities. Air France and United Airlines operate regular routes from JFK, LAX, and Chicago O'Hare. Flight times range from 7-11 hours depending on departure city.",
            "From San Francisco International Airport (SFO), Air France operates a daily direct flight AF085 departing at 3:30 PM, arriving at CDG at 11:15 AM next day. United Airlines also offers UA990 with a similar schedule. Average flight time is 10 hours 45 minutes."
        ],
        "weather": "Paris in mid-February typically experiences cool winter conditions with average daytime temperatures ranging from 8-12°C (46-54°F). Current forecast shows mostly sunny conditions with occasional cloud cover. Morning temperatures around 6°C (43°F) rising to 12°C (54°F) by afternoon. Light breeze of 8-12 km/h expected with 20% chance of precipitation. Evening temperatures dropping to 4°C (39°F). UV index moderate at 3.",
        "vector_db_results": []
    }

    save_travel_response("Paris, France", "February 11th, 2025", "February 15th, 2025", sample_research_data, "./travel_response.txt")



from typing import List, Union

from judgeval.constants import APIScorer
from judgeval.scorers.utils import (
    scorer_progress_meter,
    create_verbose_logs,
    parse_response_json,
    check_example_params
)
from judgeval.data import Example, ExampleParams
from judgeval.scorers import JudgevalScorer


required_params = [
    ExampleParams.INPUT,
    ExampleParams.ACTUAL_OUTPUT,
    ExampleParams.EXPECTED_TOOLS,
    ExampleParams.TOOLS_CALLED,
]


def get_lcs(seq1, seq2):
    m, n = len(seq1), len(seq2)
    dp = [[0] * (n + 1) for _ in range(m + 1)]

    for i in range(1, m + 1):
        for j in range(1, n + 1):
            if seq1[i - 1] == seq2[j - 1]:
                dp[i][j] = dp[i - 1][j - 1] + 1
            else:
                dp[i][j] = max(dp[i - 1][j], dp[i][j - 1])

    # Reconstruct the LCS
    lcs = []
    i, j = m, n
    while i > 0 and j > 0:
        if seq1[i - 1] == seq2[j - 1]:
            lcs.append(seq1[i - 1])
            i -= 1
            j -= 1
        elif dp[i - 1][j] > dp[i][j - 1]:
            i -= 1
        else:
            j -= 1

    return lcs[::-1]


class ToolCorrectnessScorer(JudgevalScorer):
    def __init__(
        self,
        threshold: float = 0.5,
        include_reason: bool = True,
        strict_mode: bool = False,
        verbose_mode: bool = False,
        should_exact_match: bool = False,
        should_consider_ordering: bool = False,
    ):
        super().__init__(
            score_type=APIScorer.TOOL_CORRECTNESS,
            threshold=1 if strict_mode else threshold,
            evaluation_model=None,
            include_reason=include_reason,
            async_mode=False,
            strict_mode=strict_mode,
            verbose_mode=verbose_mode
        )
        self.should_exact_match = should_exact_match
        self.should_consider_ordering = should_consider_ordering

    def measure(
        self,
        example: Example,
        _show_indicator: bool = True,
    ) -> float:
        check_example_params(example, required_params, self)

        with scorer_progress_meter(self, display_meter=_show_indicator):
            self.tools_called: List[str] = example.tools_called
            self.expected_tools: List[str] = example.expected_tools
            self.score = self._calculate_score()
            self.reason = self._generate_reason()
            self.success = self.score >= self.threshold
            self.verbose_logs = create_verbose_logs(
                self,
                steps=[
                    f"Expected Tools:\n{self.expected_tools}",
                    f"Tools Called:\n{self.tools_called}",
                    f"Score: {self.score}\nReason: {self.reason}",
                ],
            )
            return self.score

    async def a_measure(
        self, test_case: Example, _show_indicator: bool = True
    ) -> float:
        check_example_params(test_case, required_params, self)
        return self.measure(test_case, _show_indicator=_show_indicator)

    def _generate_reason(self):
        if self.should_exact_match:
            return f"{'Exact match' if self.tools_called == self.expected_tools else 'Not an exact match'}: expected {self.expected_tools}, called {self.tools_called}."

        elif self.should_consider_ordering:
            lcs = get_lcs(self.expected_tools, self.tools_called)
            missing = set(self.expected_tools) - set(self.tools_called)
            out_of_order = set(self.expected_tools) - set(lcs)

            if len(lcs) == len(self.expected_tools):
                return f"Correct ordering: all expected tools {self.expected_tools} were called in the correct order."
            else:
                issues = []
                if missing:
                    issues.append(f"missing tools {list(missing)}")
                if out_of_order:
                    issues.append(f"out-of-order tools {list(out_of_order)}")
                return f"Incorrect tool usage: {' and '.join(issues)}; expected {self.expected_tools}, called {self.tools_called}."

        else:
            used_expected = set(self.tools_called).intersection(
                set(self.expected_tools)
            )
            missing = set(self.expected_tools) - used_expected

            if len(used_expected) == len(self.expected_tools):
                return f"All expected tools {self.expected_tools} were called (order not considered)."
            else:
                return f"Incomplete tool usage: missing tools {list(missing)}; expected {self.expected_tools}, called {self.tools_called}."

    def _calculate_score(self):
        if self.should_exact_match:
            return 1.0 if self.tools_called == self.expected_tools else 0.0

        elif self.should_consider_ordering:
            longest_common_subsequence = get_lcs(
                self.expected_tools, self.tools_called
            )
            score = len(longest_common_subsequence) / len(self.expected_tools)

        else:
            used_expected_tools = set(self.tools_called).intersection(
                set(self.expected_tools)
            )
            score = len(used_expected_tools) / len(self.expected_tools)
        return 0 if self.strict_mode and score < self.threshold else score

    def _success_check(self) -> bool:
        try:
            self.success = self.score >= self.threshold
        except:
            self.success = False
        return self.success

    @property
    def __name__(self):
        return "Tool Correctness"
    


from typing import List, Optional

from judgeval.data.datasets.ground_truth import GroundTruthExample
from judgeval.data import Example


def examples_to_ground_truths(examples: List[Example]) -> List[GroundTruthExample]:
    """
    Convert a list of `Example` objects to a list of `GroundTruthExample` objects.

    Args:
        examples (List[Example]): A list of `Example` objects to convert.

    Returns:
        List[GroundTruthExample]: A list of `GroundTruthExample` objects.
    """

    if not isinstance(examples, list):
        raise TypeError("Input should be a list of `Example` objects")

    ground_truths = []
    ground_truths = []
    for e in examples:
        g_truth = {
            "input": e.input,
            "actual_output": e.actual_output,
            "expected_output": e.expected_output,
            "context": e.context,
            "retrieval_context": e.retrieval_context,
            "tools_called": e.tools_called,
            "expected_tools": e.expected_tools,
        }
        ground_truths.append(GroundTruthExample(**g_truth))
    return ground_truths


def ground_truths_to_examples(
    ground_truths: List[GroundTruthExample], 
    _alias: Optional[str] = None,
    _id: Optional[str] = None,
    ) -> List[Example]:
    """
    Converts a list of `GroundTruthExample` objects to a list of `Example` objects.

    Args:
        ground_truths (List[GroundTruthExample]): A list of `GroundTruthExample` objects to convert.
        _alias (Optional[str]): The alias of the dataset.
        _id (Optional[str]): The ID of the dataset.

    Returns:
        List[Example]: A list of `Example` objects.
    """

    if not isinstance(ground_truths, list):
        raise TypeError("Input should be a list of `GroundTruthExample` objects")

    examples = []
    for index, ground_truth in enumerate(ground_truths):
        e = Example(
            input=ground_truth.input,
            actual_output=ground_truth.actual_output,
            expected_output=ground_truth.expected_output,
            context=ground_truth.context,
            retrieval_context=ground_truth.retrieval_context,
            additional_metadata=ground_truth.additional_metadata,
            tools_called=ground_truth.tools_called,
            expected_tools=ground_truth.expected_tools,
            comments=ground_truth.comments,
            _dataset_alias=_alias,
            _dataset_id=_id,
            _dataset_rank=index,
        )
        examples.append(e)
    return examples



"""
Common Exceptions in Judgeval
"""


class MissingTestCaseParamsError(Exception):
    pass


class JudgmentAPIError(Exception):
    """
    Exception raised when an error occurs while executing a Judgment API request
    """
    
    def __init__(self, message: str):
        super().__init__(message)
        self.message = message


class InvalidJudgeModelError(Exception):
    """
    Exception raised when an invalid judge model is provided
    """
    
    def __init__(self, message: str):
        super().__init__(message)
        self.message = message




from typing import Optional, List, Union, Tuple
from pydantic import BaseModel

from judgeval.constants import APIScorer
from judgeval.judges import JudgevalJudge
from judgeval.judges.utils import create_judge
from judgeval.data import Example, ExampleParams
from judgeval.scorers import JudgevalScorer
from judgeval.scorers.utils import (
    get_or_create_event_loop,
    parse_response_json,
    scorer_progress_meter,
    create_verbose_logs,
    check_example_params,
)
from .prompts import (
    ACVerdict,
    AnswerCorrectnessTemplate,
    Statements,
    Verdicts,
    Reason,
)


required_params = [
    ExampleParams.INPUT,
    ExampleParams.ACTUAL_OUTPUT,
    ExampleParams.EXPECTED_OUTPUT,
]


class AnswerCorrectnessScorer(JudgevalScorer):
    def __init__(
        self,
        threshold: float = 0.5,
        model: Optional[Union[str, JudgevalJudge]] = None,
        include_reason: bool = True,
        async_mode: bool = True,
        strict_mode: bool = False,
        verbose_mode: bool = False
    ):
        super().__init__(
            score_type=APIScorer.ANSWER_CORRECTNESS,
            threshold=1 if strict_mode else threshold,
            evaluation_model=None,
            include_reason=include_reason,
            async_mode=async_mode,
            strict_mode=strict_mode,
            verbose_mode=verbose_mode
        )
        self.model, self.using_native_model = create_judge(model)
        self.evaluation_model = self.model.get_model_name()

    async def _a_get_statements(self, expected_output: str) -> List[str]:
        prompt = AnswerCorrectnessTemplate.deduce_statements(
            expected_output=expected_output,
        )
        if self.using_native_model:
            res = await self.model.a_generate(prompt)
            data = parse_response_json(res, self)
            return data["statements"]
        else:
            try:
                res: Statements = await self.model.a_generate(
                    prompt, schema=Statements
                )
                return res.statements
            except TypeError:
                res = await self.model.a_generate(prompt)
                data = parse_response_json(res, self)
                return data["statements"]

    def _get_statements(self, expected_output: str) -> List[str]:
        prompt = AnswerCorrectnessTemplate.deduce_statements(
            expected_output=expected_output,
        )
        if self.using_native_model:
            res = self.model.generate(prompt)
            data = parse_response_json(res, self)
            return data["statements"]
        else:
            try:
                res: Statements = self.model.generate(
                    prompt, schema=Statements
                )
                return res.statements
            except TypeError:
                res = self.model.generate(prompt)
                data = parse_response_json(res, self)
                return data["statements"]

    async def _a_get_verdicts(self, actual_output: str) -> List[ACVerdict]:
        if len(self.statements) == 0:
            return []

        prompt = AnswerCorrectnessTemplate.generate_verdicts(
            actual_output=actual_output,
            statements=self.statements,
        )

        if self.using_native_model:
            res = await self.model.a_generate(prompt)
            data = parse_response_json(res, self)
            return [ACVerdict(**item) for item in data["verdicts"]]
        else:
            try:
                res: Verdicts = await self.model.a_generate(prompt, schema=Verdicts)
                return [item for item in res.verdicts]
            except TypeError:
                res = await self.model.a_generate(prompt)
                data = parse_response_json(res, self)
                return [ACVerdict(**item) for item in data["verdicts"]]

    def _get_verdicts(self, actual_output: str) -> List[ACVerdict]:
        if len(self.statements) == 0:
            return []

        prompt = AnswerCorrectnessTemplate.generate_verdicts(
            actual_output=actual_output,
            statements=self.statements,
        )

        if self.using_native_model:
            res = self.model.generate(prompt)
            data = parse_response_json(res, self)
            return [ACVerdict(**item) for item in data["verdicts"]]
        else:
            try:
                res: Verdicts = self.model.generate(prompt, schema=Verdicts)
                return [item for item in res.verdicts]
            except TypeError:
                res = self.model.generate(prompt)
                data = parse_response_json(res, self)
                return [ACVerdict(**item) for item in data["verdicts"]]

    async def _a_get_reason(self) -> str:
        if self.include_reason is False:
            return None

        incorrect_statements: List[Tuple[str, str]] = []
        for idx, verdict in enumerate(self.verdicts):
            if verdict.verdict.strip().lower() == "no":
                incorrect_statements.append((self.statements[idx], verdict.reason))

        prompt = AnswerCorrectnessTemplate.generate_reason(
            incorrect_statements=incorrect_statements,
            score=format(self.score, ".2f"),
        )
        if self.using_native_model:
            res = await self.model.a_generate(prompt)
            data = parse_response_json(res, self)
            return data["reason"]
        else:
            try:
                res: Reason = await self.model.a_generate(
                    prompt=prompt, schema=Reason
                )
                return res.reason
            except TypeError:
                res = await self.model.a_generate(prompt)
                data = parse_response_json(res, self)
                return data["reason"]

    def _get_reason(self) -> str:
        if self.include_reason is False:
            return None

        incorrect_statements: List[Tuple[str, str]] = []
        for idx, verdict in enumerate(self.verdicts):
            if verdict.verdict.strip().lower() == "no":
                incorrect_statements.append((self.statements[idx], verdict.reason))

        prompt = AnswerCorrectnessTemplate.generate_reason(
            incorrect_statements=incorrect_statements,
            score=format(self.score, ".2f"),
        )
        if self.using_native_model:
            res = self.model.generate(prompt)
            data = parse_response_json(res, self)
            return data["reason"]
        else:
            try:
                res: Reason = self.model.generate(
                    prompt=prompt, schema=Reason
                )
                return res.reason
            except TypeError:
                res = self.model.generate(prompt)
                data = parse_response_json(res, self)
                return data["reason"]

    def _compute_score(self) -> float:
        number_of_verdicts = len(self.verdicts)
        if number_of_verdicts == 0:
            return 1

        correct_count = 0
        for verdict in self.verdicts:
            if verdict.verdict.strip().lower() == "yes":
                correct_count += 1

        score = correct_count / number_of_verdicts
        return 0 if self.strict_mode and score < self.threshold else score

    def score_example(
        self,
        example: Example,
        _show_indicator: bool = True,
    ) -> float:
        check_example_params(example, required_params, self)

        with scorer_progress_meter(self, display_meter=_show_indicator):
            try:
                if self.async_mode:
                    loop = get_or_create_event_loop()
                    loop.run_until_complete(
                        self.a_score_example(example, _show_indicator=False)
                    )
                else:
                    self.statements = self._get_statements(example.expected_output)
                    self.verdicts = self._get_verdicts(example.actual_output)
                    self.score = self._compute_score()
                    self.reason = self._get_reason()
                    self.success = self.score >= self.threshold
                    self.verbose_logs = create_verbose_logs(
                        self,
                        steps=[
                            f"Statements:\n{self.statements}",
                            f"Verdicts:\n{[v.model_dump() for v in self.verdicts]}",
                            f"Score: {self.score}\nReason: {self.reason}",
                        ],
                    )
                return self.score
            except Exception as e:
                print(f"Error in score_example for AnswerCorrectnessScorer: {e}")
                raise
    
    async def a_score_example(
        self, 
        example: Example,
        _show_indicator: bool = True,
    ) -> float:
        check_example_params(example, required_params, self)

        with scorer_progress_meter(self, async_mode=True, display_meter=_show_indicator):
            try:
                self.statements: List[str] = await self._a_get_statements(example.expected_output)
                self.verdicts: List[ACVerdict] = await self._a_get_verdicts(example.actual_output)
                self.score = self._compute_score()
                self.reason = await self._a_get_reason()
                self.success = self.score >= self.threshold
                self.verbose_logs = create_verbose_logs(
                    self,
                    steps=[
                        f"Statements:\n{self.statements}",
                        f"Verdicts:\n{[v.model_dump() for v in self.verdicts]}",
                        f"Score: {self.score}\nReason: {self.reason}",
                    ],
                )
                return self.score
            except Exception as e:
                print(f"Error in a_score_example for AnswerCorrectnessScorer: {e}")
                raise

    def _success_check(self) -> bool:
        if self.error is not None:
            self.success = False
        else:
            try:
                self.success = self.score >= self.threshold
            except:
                self.success = False
        return self.success

    @property
    def __name__(self):
        return "Answer Correctness"



"""
Code that implements a prompt-based scorer for evaluating examples.

The PromptScorer class is a base class that can be used to create custom scoring metrics using LLM prompts.
To implement a subclass of PromptScorer, you need to implement the following methods:
- build_measure_prompt(): builds the conversation prompt that is sent to the LLM judge
- build_schema(): defines the expected response schema from the LLM
- process_response(): parses the response from the LLM judge
- success_check(): determines whether the evaluation was successful

The core idea of PromptScorer is to provide a flexible way to create custom scoring metrics
by leveraging LLM judges to evaluate examples. The scorer constructs a prompt, sends it to 
the judge, and parses the structured response to determine a score.

For example, the SentimentScorer subclass uses PromptScorer to detect negative sentiment in responses
by prompting an LLM to rate the negativity on a 1-5 scale and provide a reason for the rating.

The PromptScorer supports both synchronous and asynchronous evaluation modes, includes optional
reason fields in responses, and can operate in strict mode with higher thresholds.

NOTE: When implementing build_measure_prompt and build_schema:
- The prompt should guide the LLM to generate a response matching your schema
- The schema should include "score" and optionally "reason" fields
- The score field type and range should match your scoring criteria
- The reason field provides explanatory context for the score
"""

from abc import abstractmethod
from typing import List, Optional, Union, Tuple, Any, Mapping
from pydantic import BaseModel, model_serializer, Field

from judgeval.data import Example
from judgeval.scorers import JudgevalScorer
from judgeval.scorers.utils import (scorer_progress_meter, 
                                    parse_response_json,
                                    get_or_create_event_loop,
                                    create_verbose_logs)


class ReasonScore(BaseModel):
    reason: str
    score: float


class PromptScorer(JudgevalScorer, BaseModel):
    name: str
    score_type: str
    threshold: float = Field(default=0.5)
    using_native_model: bool = Field(default=True)

    # DO NOT SET THESE FIELDS MANUALLY, THEY ARE SET BY THE SCORE_EXAMPLE METHOD
    _response: Optional[dict] = None
    _result: Optional[float] = None
    
    def __init__(
        self,
        name: str, 
        threshold: float = 0.5,
        include_reason: bool = True,
        async_mode: bool = True,
        strict_mode: bool = False,
        verbose_mode: bool = False,
    ):
        # Initialize BaseModel first
        BaseModel.__init__(
            self,
            name=name,
            score_type=name,
            threshold=1 if strict_mode else threshold,
            include_reason=include_reason,
            async_mode=async_mode,
            strict_mode=strict_mode,
            verbose_mode=verbose_mode,
        )
        # Then initialize CustomScorer
        JudgevalScorer.__init__(
            self,
            score_type=name,
            threshold=1 if strict_mode else threshold,
            include_reason=include_reason,
            async_mode=async_mode,
            strict_mode=strict_mode,
            verbose_mode=verbose_mode,
        )

    def score_example(
            self, 
            example: Example, 
            _show_indicator: bool = True
            ) -> float:
        """
        Synchronous method for scoring an example using the prompt criteria.
        """
        with scorer_progress_meter(self, display_meter=_show_indicator):
            if self.async_mode:
                loop = get_or_create_event_loop()
                loop.run_until_complete(
                    self.a_score_example(example, _show_indicator=False)
                )
            else:
                result, reason = self.evaluate(example)
                self.reason = reason
                self._result = result
                self.verbose_logs = create_verbose_logs(
                    self,
                    steps=[
                        f"Results: {self._result}\nReason: {self.reason}",
                    ],
                )
                return result

    async def a_score_example(
            self,
            example: Example,
            _show_indicator: bool = True,
            ) -> float: 
        """
        Async method for scoring an example using the prompt criteria.
        """
        with scorer_progress_meter(self, display_meter=_show_indicator):
            result, reason = await self.a_evaluate(example)
            self.reason = reason
            self._result = result
            self.verbose_logs = create_verbose_logs(
                self,
                steps=[
                    f"Results: {self._result}\nReason: {self.reason}",
                ],
            )
            return result
    
    def evaluate(self, example: Example) -> Tuple[Any, str]:
        """
        Synchronous helper method for evaluating an example using the prompt criteria.

        Builds a custom prompt using `build_measure_prompt` and sends it to the judge model 
        for evaluation. The result is then parsed as JSON and returned.

        NOTE: It is assumed that the model response will be JSON and contain a "score" and "reason" field.
        """
        prompt = self._build_measure_prompt(example)
        if self.using_native_model:
            res = self.model.generate(prompt)
            response = parse_response_json(res, self)
            result, reason = self._process_response(response)
            return result, reason
        else:
            raise NotImplementedError("Non-native judge models are not supported in synchronous mode yet.")

    async def a_evaluate(self, example: Example) -> Tuple[Any, str]:
        """
        Asynchronous helper method for evaluating an example using the prompt criteria.

        Builds a custom prompt using `build_measure_prompt` and sends it to the judge model 
        for evaluation. The result is then parsed as JSON and returned.

        NOTE: It is assumed that the model response will be JSON and contain a "score" and "reason" field.
        """
        judge_prompt = self._build_measure_prompt(example)
        schema = self._build_schema()
        prompt = self._enforce_prompt_format(judge_prompt=judge_prompt, schema=schema)
        if self.using_native_model:
            res = await self.model.a_generate(prompt)
            response = parse_response_json(res, self)
            self._response = response

            result, reason = self._process_response(response)
            self.score = result
            self.reason = reason
            self._response = response
            return result, reason
        else:
            raise NotImplementedError("Non-native judge models are not supported in async mode yet.")

    # TODO: can we make this take *args and **kwargs? How does that work with a_evaluate() since we'd have to pass the same args
    @abstractmethod
    def _build_measure_prompt(self, example: Example) -> List[dict]:
        # builds the prompt that is sent to the model inside of the `score_example()` method
        # returns either a string prompt or a conversation prompt of the form [{"role": "system", "content": "..."}, ...]

        """
        This function creates the prompt that the judge model uses to evaluate examples.

        The prompt is typically a set of instructions that the judge model uses to evaluate the example.

        This function returns a conversation prompt of the form 
        [{"role": "system", "content": "..."}, {"role": "user", "content": "..."}]

        A basic version of implementing this function could be as follows:
        SYSTEM_ROLE = ...
        return [
            {"role": "system", "content": SYSTEM_ROLE},
            {"role": "user", "content": f"Response: {example.actual_output}\n\nYour judgment: "}
        ]
        """
        pass
    
    # TODO: does this need to take *args and **kwargs? How does that work with a_evaluate() since we'd have to pass the same args
    @abstractmethod
    def _build_schema(self) -> dict:
        """
        This function returns a dictionary that represents the schema of the JSON response that the judge model should return.

        The keys of the dictionary are the expected keys in the response, and the values are the types of the corresponding values.

        Example: If you want to have the judge model return a score and a reason, you would write:
        return {"score": int, "reason": str}
        """
        pass
    
    def _enforce_prompt_format(self, judge_prompt: List[dict], schema: dict):
        """
        Formats the final prompt to the judge model.

        This function takes a list of dictionaries (`judge_prompt`) and a schema dictionary (`schema`), 
        and appends a schema enforcement prompt to the content of the first dictionary in the list, which is assumed to be the system prompt. 
        The schema enforcement prompt instructs the judge model to provide its response in a specific JSON format.

        Args:
            judge_prompt (List[dict]): A list of dictionaries representing the judge prompt. 
                                       Each dictionary should contain a "content" key.
            schema (dict): A dictionary representing the schema. The keys are the expected keys in the response, 
                           and the values are the types of the corresponding values.

        Returns:
            List[dict]: The modified judge prompt with the schema enforcement prompt appended to the content 
                        of the first dictionary.

        Raises:
            TypeError: If `judge_prompt` is not a list of dictionaries.

        Example:
            judge_prompt = [{"content": "Please evaluate the following:"}]
            schema = {"score": int, "comments": str}
            formatted_prompt = format_measure_prompt(judge_prompt, schema)
            # formatted_prompt[0]["content"] will include the schema enforcement prompt
        """
        SCHEMA_ENFORCEMENT_PROMPT = "\n\nPlease provide your response in the following JSON format: {"
        if isinstance(judge_prompt, list) and all(isinstance(item, dict) for item in judge_prompt):
            # create formatting string for schema enforcement
            # schema is a map between key and type of the value 
            for key, key_type in schema.items():
                SCHEMA_ENFORCEMENT_PROMPT += f'"{key}": <{key}> ({key_type.__name__}), '
            SCHEMA_ENFORCEMENT_PROMPT = SCHEMA_ENFORCEMENT_PROMPT[:-2] + "}"  # remove trailing comma and space
            judge_prompt[0]["content"] += SCHEMA_ENFORCEMENT_PROMPT
            return judge_prompt
        else:
            raise TypeError(f"Prompt must be a list of dictionaries. Got {type(judge_prompt)} instead.")

    @abstractmethod 
    def _process_response(self, response: dict):
        """
        Customizable method for processing the response from the judge model.

        You can add any additional logic to parse the JSON response here and return the result and reason for decision.

        If you don't need a reason for the decision, you can simply return (score, None).

        Example:
        score = response["score"]
        reason = response["reason"]
        return score, reason
        """
        pass

    @abstractmethod
    def _success_check(self, **kwargs) -> bool:
        """
        Determines whether or not the PromptScorer should consider the evaluation of a single example successful.
        """
        pass
    
    @property
    def __name__(self):
        return self.name


class ClassifierScorer(PromptScorer):

    """
    This is a PromptScorer that takes 
    1. a system role that may involve the Example object
    2. options for scores on the example
    
    and uses a judge to execute the evaluation from the system role and classify into one of the options

    ex:
    system_role = "You are a judge that evaluates whether the response is positive or negative. The response is: {example.actual_output}"
    options = {"positive": 1, "negative": 0}
    """

    conversation: List[dict]
    options: Mapping[str, float]
    
    def __init__(self, name: str, slug: str, conversation: List[dict], options: Mapping[str, float], 
                 threshold: float = 0.5, include_reason: bool = True, 
                 async_mode: bool = True, strict_mode: bool = False, verbose_mode: bool = False):
        # Initialize BaseModel first with all fields
        BaseModel.__init__(
            self,
            name=name,
            slug=slug,
            score_type=name,
            conversation=conversation,
            options=options,
            threshold=threshold,
            include_reason=include_reason,
            async_mode=async_mode,
            strict_mode=strict_mode,
            verbose_mode=verbose_mode,
        )
        # Then initialize CustomScorer
        JudgevalScorer.__init__(
            self,
            score_type=name,
            threshold=threshold,
            include_reason=include_reason,
            async_mode=async_mode,
            strict_mode=strict_mode,
            verbose_mode=verbose_mode,
        )

    def _build_measure_prompt(self, example: Example) -> List[dict]:
        """
        Builds the measure prompt for the classifier scorer.

        Args:
            example (Example): The example to build the prompt for

        Returns:
            List[dict]: The measure prompt for the classifier scorer
        """
        replacement_words = {
            "{{actual_output}}": example.actual_output,
            "{{expected_output}}": example.expected_output,
            "{{context}}": example.context,
            "{{retrieval_context}}": example.retrieval_context,
            "{{tools_called}}": example.tools_called,
            "{{expected_tools}}": example.expected_tools,
        }
        # Make a copy of the conversation to avoid modifying the original
        conversation_copy = [dict(message) for message in self.conversation]
        
        # Only replace if double brackets are found in the content
        for message in conversation_copy:
            content = message["content"]
            if "{{" in content:
                for key, value in replacement_words.items():
                    if key in content:
                        message["content"] = content.replace(key, str(value))
        return conversation_copy

    def _build_schema(self) -> dict:
        return self.options
    
    def _enforce_prompt_format(self, judge_prompt: List[dict], schema: dict) -> List[dict]:
        """
        Enforces the judge model to choose an option from the schema.

        We want the model to choose an option from the schema and a reason for the choice.
        """
        options = list(schema.keys())
        options_str = ", ".join(options)
        
        system_role = judge_prompt[0]["content"]
        system_role += (
            f"\n\nYou must choose one of the following options: {options_str}. "
            "Format your response as a JSON object with two fields:\n"
            "1. 'choice': Your selected option (must be one of the provided choices)\n"
            "2. 'reason': A brief explanation for why you made this choice\n\n"
            "Example response format:\n"
            "{\n"
            '    "choice": "<one of the valid options>",\n'
            '    "reason": "<your explanation>"\n'
            "}"
        )
        
        judge_prompt[0]["content"] = system_role
        return judge_prompt

    def _process_response(self, response: dict) -> Tuple[float, str]:
        choice = response.get("choice")
        if choice not in self.options:
            raise ValueError(f"Invalid choice: {choice}. Expected one of: {self.options.keys()}")
        reason = response.get("reason", "No reason could be found in model response.")
        return self.options[choice], reason

    def _success_check(self, **kwargs) -> bool:
        return self.score >= self.threshold
    
    def update_name(self, name: str):
        """
        Updates the name of the scorer.
        """
        self.name = name
        
    def update_threshold(self, threshold: float):
        """
        Updates the threshold of the scorer.
        """
        self.threshold = threshold
    
    def update_conversation(self, conversation: List[dict]):
        """
        Updates the conversation with the new conversation.
        
        Sample conversation:
        [{'role': 'system', 'content': "Did the chatbot answer the user's question in a kind way?: {{actual_output}}."}]
        """
        self.conversation = conversation
        
    def update_options(self, options: Mapping[str, float]):
        """
        Updates the options with the new options.
        
        Sample options:
        {"yes": 1, "no": 0}
        """
        self.options = options

    def __str__(self):
        return f"ClassifierScorer(name={self.name}, slug={self.slug}, conversation={self.conversation}, threshold={self.threshold}, options={self.options})"

    @model_serializer
    def serialize_model(self) -> dict:
        """
        Defines how the ClassifierScorer should be serialized when model_dump() is called.
        """
        return {
            "name": self.name,
            "score_type": self.score_type,
            "conversation": self.conversation,
            "options": self.options,
            "threshold": self.threshold,
            "include_reason": self.include_reason,
            "async_mode": self.async_mode,
            "strict_mode": self.strict_mode,
            "verbose_mode": self.verbose_mode,
        }
